#   Copyright 2017 The OpenFermion Developers
#   Modifications copyright 2022 Zapata Computing, Inc. for compatibility reasons.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""This module provides generic tools for classes in orquestra.quantum.operators"""

import numpy
from scipy.sparse import spmatrix

from .._pauli_operators import PauliSum, PauliTerm
from .config import EQ_TOLERANCE


def hermitian_conjugated(operator):
    """Return Hermitian conjugate of operator."""
    # Handle PauliSum
    if isinstance(operator, PauliSum):
        conjugate_operator = PauliSum()
        for term in operator.terms:
            conjugate_operator += term.copy(term.coefficient.conjugate())

    # Handle PauliTerm
    elif isinstance(operator, PauliTerm):
        conjugate_operator = operator.copy(operator.coefficient.conjugate())

    # Handle sparse matrix
    elif isinstance(operator, spmatrix):
        conjugate_operator = operator.getH()

    # Handle numpy array
    elif isinstance(operator, numpy.ndarray):
        conjugate_operator = operator.T.conj()

    # Unsupported type
    else:
        raise TypeError(
            "Taking the hermitian conjugate of a {} is not "
            "supported.".format(type(operator).__name__)
        )

    return conjugate_operator


def is_hermitian(operator):
    """Test if operator is Hermitian."""
    # Handle QubitOperator
    if isinstance(operator, (PauliSum, PauliTerm)):
        return operator == hermitian_conjugated(operator)

    # Handle sparse matrix
    elif isinstance(operator, spmatrix):
        difference = operator - hermitian_conjugated(operator)
        discrepancy = 0.0
        if difference.nnz:
            discrepancy = max(abs(difference.data))
        return discrepancy < EQ_TOLERANCE

    # Handle numpy array
    elif isinstance(operator, numpy.ndarray):
        difference = operator - hermitian_conjugated(operator)
        discrepancy = numpy.amax(abs(difference))
        return discrepancy < EQ_TOLERANCE

    # Unsupported type
    else:
        raise TypeError(
            "Checking whether a {} is hermitian is not "
            "supported.".format(type(operator).__name__)
        )



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
from typing import List

from orquestra.quantum.api.estimation import EstimationTask
from orquestra.quantum.api.estimator_contract import ESTIMATOR_CONTRACTS
from orquestra.quantum.api.wavefunction_simulator import WavefunctionSimulator
from orquestra.quantum.estimation import calculate_exact_expectation_values
from orquestra.quantum.measurements import ExpectationValues

# This function will be used as a mock estimator
_good_estimator = calculate_exact_expectation_values


def test_each_task_returns_one_expecation_value_test():
    def malicious_calculate_expectation_values(
        runner: WavefunctionSimulator, estimation_tasks: List[EstimationTask]
    ) -> List[ExpectationValues]:
        correct_output = calculate_exact_expectation_values(
            runner=runner, estimation_tasks=estimation_tasks
        )

        return correct_output[:-1]

    bad_estimator = malicious_calculate_expectation_values

    assert ESTIMATOR_CONTRACTS[0](_good_estimator)
    assert not ESTIMATOR_CONTRACTS[0](bad_estimator)


def test_order_of_outputs_matches_order_of_inputs_test():
    def malicious_calculate_expectation_values(
        runner: WavefunctionSimulator, estimation_tasks: List[EstimationTask]
    ) -> List[ExpectationValues]:
        correct_output = calculate_exact_expectation_values(
            runner=runner, estimation_tasks=estimation_tasks
        )

        correct_output[0], correct_output[-1] = correct_output[-1], correct_output[0]

        return correct_output

    bad_estimator = malicious_calculate_expectation_values

    assert ESTIMATOR_CONTRACTS[1](_good_estimator)
    assert not ESTIMATOR_CONTRACTS[1](bad_estimator)


def test_expectation_value_includes_coefficients_test():
    def malicious_calculate_expectation_values(
        runner: WavefunctionSimulator, estimation_tasks: List[EstimationTask]
    ) -> List[ExpectationValues]:
        correct_output = calculate_exact_expectation_values(
            runner=runner, estimation_tasks=estimation_tasks
        )

        correct_output[1] = correct_output[0]

        return correct_output

    bad_estimator = malicious_calculate_expectation_values

    assert ESTIMATOR_CONTRACTS[2](_good_estimator)
    assert not ESTIMATOR_CONTRACTS[2](bad_estimator)


def test_constant_terms_are_included_in_output_test():
    def malicious_calculate_expectation_values(
        runner: WavefunctionSimulator, estimation_tasks: List[EstimationTask]
    ) -> List[ExpectationValues]:
        correct_output = calculate_exact_expectation_values(
            runner=runner, estimation_tasks=estimation_tasks
        )

        correct_output[1] = correct_output[0]

        return correct_output

    bad_estimator = malicious_calculate_expectation_values

    assert ESTIMATOR_CONTRACTS[3](_good_estimator)
    assert not ESTIMATOR_CONTRACTS[3](bad_estimator)



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
import json
import math
import warnings
from io import StringIO
from itertools import product
from sys import float_info
from unittest import mock

import numpy as np
import pytest

from orquestra.quantum.distributions._measurement_outcome_distribution import (
    MeasurementOutcomeDistribution,
    _are_keys_non_negative_integer_tuples,
    _is_key_length_fixed,
    _is_non_negative,
    change_tuple_dict_keys_to_comma_separated_integers,
    create_bitstring_distribution_from_probability_distribution,
    evaluate_distribution_distance,
    is_measurement_outcome_distribution,
    is_normalized,
    load_measurement_outcome_distribution,
    load_measurement_outcome_distributions,
    normalize_measurement_outcome_distribution,
    preprocess_distibution_dict,
    save_measurement_outcome_distribution,
    save_measurement_outcome_distributions,
)


class TestVerifiersAndValidators:
    @pytest.mark.parametrize(
        "validator,positive_case",
        [
            (_is_non_negative, {i: i + 1 for i in range(10)}),
            (
                _is_key_length_fixed,
                {("a", "b", "c"): 3, (1, 0, 0): 2, ("w", "w", "w"): 1},
            ),
            (
                _are_keys_non_negative_integer_tuples,
                {(1, 0, 0, 0, 0, 1): 3, (1, 99): 2, (0, 45, 36, 1): 1},
            ),
            (
                is_measurement_outcome_distribution,
                {(1, 0, 0): 3, (1, 1, 0): 2, (0, 1, 0): 1},
            ),
            (
                is_measurement_outcome_distribution,
                preprocess_distibution_dict(
                    {
                        "110": 0.5,
                        (1, 0, 0): 0.5,
                    }
                ),
            ),
            (
                is_measurement_outcome_distribution,
                preprocess_distibution_dict(
                    {
                        "1,1,0": 0.5,
                        (1, 0, 0): 0.5,
                    }
                ),
            ),
        ],
    )
    def test_validator_returns_true_for_positive_case(self, validator, positive_case):
        assert validator(positive_case)

    @pytest.mark.parametrize(
        "validator,negative_case",
        [
            (_is_non_negative, {i: -i for i in range(10)}),
            (_is_non_negative, {0: -1, 1: 2, 3: 0}),
            (_is_key_length_fixed, {("a"): 3, (1, 0): 2, ("w", "w", "w"): 1}),
            (
                _are_keys_non_negative_integer_tuples,
                {("a", "b", "c"): 3, (1, 0, 0): 2, ("w", "w", "w"): 1},
            ),
            (
                is_measurement_outcome_distribution,
                {(1, 0, 0, 0, 0, 1): 3, (1, 0): 2, (0, 1, 0, 1): 1},
            ),
            (
                is_measurement_outcome_distribution,
                {("a", "b", "c"): 3, (1, 0, 0): 2, ("w", "w", "w"): 1},
            ),
            (
                is_measurement_outcome_distribution,
                {
                    "abc": 0.5,
                    (1, 0, 0): 0.5,
                },
            ),
            (
                is_measurement_outcome_distribution,
                {
                    "a,b,c": 0.5,
                    (1, 0, 0): 0.5,
                },
            ),
        ],
    )
    def test_validator_returns_false_for_negative_case(self, validator, negative_case):
        assert not validator(negative_case)

    @pytest.mark.parametrize(
        "distribution",
        [
            {(0, 0, 0): 0.1, (1, 1, 1): 0.9},
            {(0, 1, 0): 0.3, (0, 0, 0): 0.2, (1, 1, 1): 0.5},
            {
                (0, 1, 0): 0.3,
                (0, 0, 0): 0.2,
                (1, 1, 1): 0.1,
                (1, 0, 0): 0.4,
            },
        ],
    )
    def test_distributions_with_probabilities_summing_to_one_are_normalized(
        self,
        distribution,
    ):
        assert is_normalized(distribution)

    @pytest.mark.parametrize(
        "distribution",
        [
            {(0, 0, 0): 0.1, (1, 1, 1): 9},
            {(0, 0, 0): 2, (1, 1, 1): 0.9},
            {(0, 0, 0): 1e-3, (1, 1, 1): 0, (1, 0, 0): 100},
        ],
    )
    def test_distributions_with_probabilities_not_summing_to_one_are_not_normalized(
        self,
        distribution,
    ):
        assert not is_normalized(distribution)

    def test_preprocessor_raises_error_for_invalid_input(self):
        with pytest.raises(RuntimeError):
            preprocess_distibution_dict({1.35: 1.0})


class TestInitializations:
    @pytest.mark.parametrize(
        "prob_dist,expected_dist",
        [
            (
                np.asarray([0.25, 0, 0.5, 0.25]),
                MeasurementOutcomeDistribution(
                    {(0, 0): 0.25, (1, 0): 0.5, (0, 1): 0.0, (1, 1): 0.25}
                ),
            ),
            (
                np.ones(2**5) / 2**5,
                MeasurementOutcomeDistribution(
                    {tup: 1 / 2**5 for tup in product([0, 1], repeat=5)}
                ),
            ),
        ],
    )
    def test_constructs_correct_bitstring_distribution_from_probability_distribution(
        self, prob_dist, expected_dist
    ):
        """Probability distributions is converted to matching bitstring distributions.

        Bitstring distributions constructed from probability distribution should have:
        - keys equal to binary representation of consecutive natural numbers,
        - values corresponding to original probabilities.
        """
        bitstring_dist = create_bitstring_distribution_from_probability_distribution(
            prob_dist
        )
        assert bitstring_dist.distribution_dict == expected_dist.distribution_dict
        assert (
            bitstring_dist.get_number_of_subsystems()
            == expected_dist.get_number_of_subsystems()
        )

    @pytest.mark.parametrize(
        "distribution",
        [
            {(0, 0, 0): 0.1, (1, 1, 1): 0.9},
            {(0, 1, 0): 0.3, (1, 1, 1): 0.9},
            {(0, 0, 0): 2, (1, 1, 1): 0.9},
            {(0, 0, 0): 2, (1, 1, 1): 4.9},
            {(0, 0, 0): 0.2, (1, 1, 1): 9},
            {(0, 0, 0): 1e-3, (1, 1, 1): 0},
        ],
    )
    def test_distribution_gets_normalized_by_default(self, distribution):
        distribution = MeasurementOutcomeDistribution(distribution)
        assert is_normalized(distribution.distribution_dict)

    def test_original_dict_is_kept_if_normalization_isnt_requested_and_warns(
        self,
    ):
        distribution_dict = {(0, 0, 0): 0.1, (1, 1, 1): 9}
        with pytest.warns(UserWarning):
            distribution = MeasurementOutcomeDistribution(
                {(0, 0, 0): 0.1, (1, 1, 1): 9}, normalize=False
            )
        assert distribution.distribution_dict == distribution_dict

    @pytest.mark.parametrize(
        "distribution,num_qubits",
        [
            (MeasurementOutcomeDistribution({(0, 0): 0.1, (1, 1): 0.9}), 2),
            (MeasurementOutcomeDistribution({(0, 0, 0): 0.2, (1, 1, 1): 0.8}), 3),
            (MeasurementOutcomeDistribution({(0, 0, 0, 0): 1e-3, (1, 1, 1, 1): 0}), 4),
        ],
    )
    def test_number_of_qubits_in_distribution_equals_length_of_keys(
        self, distribution, num_qubits
    ):
        assert distribution.get_number_of_subsystems() == num_qubits

    def test_constructor_invalid_distribution_throws_error(self):
        with pytest.raises(RuntimeError):
            MeasurementOutcomeDistribution({(0, 1, 0): 0.1, (1,): 0.9})


def test_repr_function_returns_expected_string():
    dictionary = {(0,): 0.1, (1,): 0.9}
    dist = MeasurementOutcomeDistribution(dictionary)

    assert dist.__repr__() == f"MeasurementOutcomeDistribution(input={dictionary})"


class TestNormalization:
    def test_normalizing_normalized_dict_does_nothing(self):
        assert normalize_measurement_outcome_distribution({(0,): 1.0}) == {(0,): 1.0}

    @pytest.mark.parametrize(
        "dist",
        [
            {(0, 0, 0): 0.1, (1, 1, 1): 9},
            {(0, 0, 0): 2, (1, 1, 1): 0.9},
            {(0, 0, 0): 1e-3, (1, 1, 1): 0, (1, 0, 0): 100},
        ],
    )
    def test_normalizing_distribution_gives_normalized_distribution(self, dist):
        assert not is_normalized(dist)
        normalize_measurement_outcome_distribution(dist)
        assert is_normalized(dist)

    @pytest.mark.parametrize(
        "dist,expected_error_content",
        [
            ({(0, 0, 0): 0.0, (1, 1, 1): 0.0}, "all zero values"),
            ({(0, 0, 0): float_info.min / 2}, "too small values"),
        ],
    )
    def test_normalizing_distribution_raises_error_for_values_with_invalid_norm(
        self, dist, expected_error_content
    ):
        with pytest.raises(ValueError) as error:
            normalize_measurement_outcome_distribution(dist)

        assert expected_error_content in error.value.args[0]


def test_passed_measure_is_used_for_evaluating_distribution_distance():
    """Evaluating distance distribution uses distance measure passed as an argument."""
    target_distribution = MeasurementOutcomeDistribution({(0,): 10, (1,): 5})
    measured_distribution = MeasurementOutcomeDistribution({(0,): 10, (1,): 5})
    distance_function = mock.Mock()

    distance = evaluate_distribution_distance(
        target_distribution, measured_distribution, distance_function
    )

    distance_function.assert_called_once_with(
        target_distribution, measured_distribution
    )
    assert distance == distance_function.return_value


class TestSavingDistributions:
    @pytest.fixture
    def mock_open(self):
        mock_open = mock.mock_open()
        with mock.patch(
            "orquestra.quantum.distributions._measurement_outcome_distribution.open",
            mock_open,
        ):
            yield mock_open

    def test_saving_distribution_opens_file_for_writing_using_context_manager(
        self,
        mock_open,
    ):
        distribution = MeasurementOutcomeDistribution({(0, 0, 0): 0.1, (1, 1, 1): 0.9})
        save_measurement_outcome_distribution(
            distribution, "/some/path/to/distribution.json"
        )

        mock_open.assert_called_once_with("/some/path/to/distribution.json", "w")
        mock_open().__enter__.assert_called_once()
        mock_open().__exit__.assert_called_once()

    def test_saving_distributions_opens_file_for_writing_using_context_manager(
        self,
        mock_open,
    ):
        distributions = [
            MeasurementOutcomeDistribution({(0, 0, 0): 0.1, (1, 1, 1): 0.9}),
            MeasurementOutcomeDistribution(
                {(0, 1, 0, 0, 0): 0.5, (1, 0, 1, 1, 0): 0.5}
            ),
        ]
        save_measurement_outcome_distributions(
            distributions, "/some/path/to/distribution/set.json"
        )

        mock_open.assert_called_once_with("/some/path/to/distribution/set.json", "w")
        mock_open().__enter__.assert_called_once()
        mock_open().__exit__.assert_called_once()

    def test_saving_distribution_writes_correct_json_data_to_file(self, mock_open):
        """Saving distribution writes correct json dictionary to file."""
        distribution = MeasurementOutcomeDistribution({(0, 0, 0): 0.1, (1, 1, 1): 0.9})

        preprocessed_dict = change_tuple_dict_keys_to_comma_separated_integers(
            distribution.distribution_dict
        )

        expected_dict = {
            "measurement_outcome_distribution": preprocessed_dict,
        }

        save_measurement_outcome_distribution(
            distribution, "/some/path/to/distribution.json"
        )

        written_data = mock_open().__enter__().write.call_args[0][0]
        assert json.loads(written_data) == expected_dict

    def test_saving_distributions_writes_correct_json_data_to_file(self, mock_open):
        distributions = [
            MeasurementOutcomeDistribution({(0, 0, 0): 0.1, (1, 1, 1): 0.9}),
            MeasurementOutcomeDistribution(
                {(0, 1, 0, 0, 0): 0.5, (1, 0, 1, 1, 0): 0.5}
            ),
        ]

        expected_dict = {
            "measurement_outcome_distribution": [
                change_tuple_dict_keys_to_comma_separated_integers(
                    distribution.distribution_dict
                )
                for distribution in distributions
            ],
        }

        save_measurement_outcome_distributions(
            distributions, "/some/path/to/distribution/set.json"
        )

        written_data = mock_open().__enter__().write.call_args[0][0]
        assert json.loads(written_data) == expected_dict

    def test_saved_distribution_can_be_loaded_back(self, mock_open):
        fake_file = StringIO()
        mock_open().__enter__.return_value = fake_file
        dist = MeasurementOutcomeDistribution({(0, 0, 0): 0.1, (1, 1, 1): 0.9})

        save_measurement_outcome_distribution(dist, "distribution.json")
        fake_file.seek(0)

        loaded_dist = load_measurement_outcome_distribution(fake_file)
        assert all(
            math.isclose(
                dist.distribution_dict[key], loaded_dist.distribution_dict[key]
            )
            for key in dist.distribution_dict.keys()
        )

        assert dist.distribution_dict.keys() == loaded_dist.distribution_dict.keys()

    def test_saved_distributions_can_be_loaded(self, mock_open):
        fake_file = StringIO()
        mock_open().__enter__.return_value = fake_file
        distributions = [
            MeasurementOutcomeDistribution({(0, 0, 0): 0.1, (1, 1, 1): 0.9}),
            MeasurementOutcomeDistribution(
                {(0, 1, 0, 0, 0): 0.5, (1, 0, 1, 1, 0): 0.5}
            ),
        ]

        save_measurement_outcome_distributions(distributions, "distributions.json")
        fake_file.seek(0)

        loaded_distributions = load_measurement_outcome_distributions(fake_file)
        assert all(
            (
                math.isclose(
                    distribution.distribution_dict[key],
                    loaded_distribution.distribution_dict[key],
                )
                for key in distribution.distribution_dict.keys()
            )
            for distribution, loaded_distribution in zip(
                distributions, loaded_distributions
            )
        )

        assert all(
            distribution.distribution_dict.keys()
            == loaded_distribution.distribution_dict.keys()
            for distribution, loaded_distribution in zip(
                distributions, loaded_distributions
            )
        )


class TestSubdistribution:
    @pytest.mark.parametrize(
        "input_distribution,active_qubits,target_dict",
        [
            (
                MeasurementOutcomeDistribution(
                    {"001": 0.5, "010": 0.3, "011": 0.2},
                ),
                [0, 1, 2],
                {(0, 0, 1): 0.5, (0, 1, 0): 0.3, (0, 1, 1): 0.2},
            ),
            (
                MeasurementOutcomeDistribution({"001": 50, "010": 45, "011": 5}),
                [0, 2],
                {(0, 1): 0.55, (0, 0): 0.45},
            ),
            (
                MeasurementOutcomeDistribution(
                    {"001": 0.5, "010": 0.3, "011": 0.2},
                ),
                [1, 0, 2],
                {(0, 0, 1): 0.5, (1, 0, 0): 0.3, (1, 0, 1): 0.2},
            ),
            (
                MeasurementOutcomeDistribution(
                    {"001": 0.5, "010": 0.3, "011": 0.2},
                ),
                [1, 0],
                {(0, 0): 0.5, (1, 0): 0.5},
            ),
            (
                MeasurementOutcomeDistribution(
                    {"001": 0.5, "010": 0.3, "011": 0.2},
                ),
                [2, 1, 0],
                {(1, 0, 0): 0.5, (0, 1, 0): 0.3, (1, 1, 0): 0.2},
            ),
        ],
    )
    def test_simple_subdistributions(
        self, input_distribution, active_qubits, target_dict
    ):
        assert (
            input_distribution.subdistribution(active_qubits).distribution_dict
            == target_dict
        )

    @pytest.mark.parametrize(
        "input_distribution,active_qubits",
        [
            (
                MeasurementOutcomeDistribution({"001": 100, "010": 101, "011": 7}),
                [0, 1, 1],
            ),
            (
                MeasurementOutcomeDistribution({"001": 100, "010": 101, "011": 7}),
                [0, 2, 3],
            ),
            (
                MeasurementOutcomeDistribution({"001": 100, "010": 101, "011": 7}),
                [],
            ),
        ],
    )
    def test_if_subdistribution_raised_error_for_invalid_input(
        self, input_distribution, active_qubits
    ):
        with pytest.raises(ValueError):
            assert input_distribution.subdistribution(active_qubits)



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
import json
from functools import lru_cache
from math import log2
from typing import Any, Dict, List, Optional, Sequence, Set, Tuple, Union
from warnings import warn

import numpy as np
from sympy import Matrix, Symbol

from .typing import AnyPath, LoadSource, ParameterizedVector
from .utils import (
    convert_array_to_dict,
    convert_bitstrings_to_tuples,
    convert_dict_to_array,
    ensure_open,
)


def _is_number(possible_number):
    try:
        complex(possible_number)
        return True
    except Exception:
        return False


def _cast_sympy_matrix_to_numpy(sympy_matrix, complex=False):
    new_type = np.complex128 if complex else np.float64

    try:
        return np.array(sympy_matrix, dtype=new_type).flatten()
    except TypeError:
        return np.array(sympy_matrix, dtype=object).flatten()


def _get_next_number_with_same_hamming_weight(val):
    # Copied from:
    # http://graphics.stanford.edu/~seander/bithacks.html#NextBitPermutation
    t = (val | (val - 1)) + 1
    return t | ((((t & -t) // (val & -val)) >> 1) - 1)


def _most_significant_set_bit(val):
    bin_string = bin(val)
    return len(bin_string) - 2


class Wavefunction:
    """
    A simple wavefunction data structure that can
    be used to calculate amplitudes of quantum states.

    Args:
        amplitude_vector: the initial amplitudes of the system,
            can either be a NumPy ndarray or a SymPy Matrix
    """

    def __init__(self, amplitude_vector: ParameterizedVector) -> None:
        if bin(len(amplitude_vector)).count("1") != 1:
            raise ValueError(
                "Provided wavefunction does not have a size of a power of 2."
            )

        try:
            self._amplitude_vector = np.asarray(amplitude_vector, dtype=complex)
        except TypeError:
            self._amplitude_vector = Matrix(amplitude_vector)

        self._check_normalization(self._amplitude_vector)

    @property
    def amplitudes(self) -> Union[np.ndarray, Matrix]:
        if self.free_symbols:
            return _cast_sympy_matrix_to_numpy(self._amplitude_vector, complex=True)

        return self._amplitude_vector

    @property
    def n_qubits(self):
        return int(log2(len(self)))

    @property
    def free_symbols(self) -> Set[Symbol]:
        return getattr(self._amplitude_vector, "free_symbols", set())

    @staticmethod
    def _check_normalization(arr: ParameterizedVector):

        if (
            isinstance(arr, np.ndarray)
            or isinstance(arr, Matrix)
            and not arr.free_symbols
        ):
            probs_of_ground_entries = np.sum(np.abs(arr) ** 2)

            if not np.isclose(probs_of_ground_entries, 1.0):
                raise ValueError("Vector does not result in a unit probability.")
        else:
            numbers = np.array(
                [elem for elem in arr if _is_number(elem)], dtype=np.complex128
            )
            probs_of_ground_entries = np.sum(np.abs(numbers) ** 2)

            if probs_of_ground_entries > 1.0:
                raise ValueError(
                    "Ground entries in vector already exceeding probability of 1.0!"
                )

    def __len__(self) -> int:
        return len(self._amplitude_vector)

    def __iter__(self):
        return iter(self._amplitude_vector)

    def __getitem__(self, idx):
        return self._amplitude_vector[idx]

    def __setitem__(self, idx, val):
        old_val = self._amplitude_vector[idx]
        self._amplitude_vector[idx] = val

        try:
            self._check_normalization(self._amplitude_vector)
        except ValueError:
            self._amplitude_vector[idx] = old_val

            raise ValueError("This assignment violates probability unity.")

    def __str__(self) -> str:
        cast_wf = _cast_sympy_matrix_to_numpy(self._amplitude_vector, complex=True)
        return f"Wavefunction({cast_wf})"

    def __eq__(self, other) -> bool:
        if not isinstance(other, Wavefunction):
            return False

        return np.array_equal(self.amplitudes, other.amplitudes)

    @staticmethod
    def zero_state(n_qubits: int) -> "Wavefunction":
        if not isinstance(n_qubits, int):
            warn(
                f"Non-integer value {n_qubits} passed as number of qubits! "
                "Will be cast to integer."
            )
            n_qubits = int(n_qubits)

        if n_qubits <= 0:
            raise ValueError(f"Invalid number of qubits in system. Got {n_qubits}.")

        np_arr = np.zeros(2**n_qubits, dtype=np.complex128)
        np_arr[0] = 1.0
        return Wavefunction(np_arr)

    @staticmethod
    def dicke_state(n_qubits: int, hamming_weight: int) -> "Wavefunction":
        initial_wf = Wavefunction.zero_state(n_qubits)

        if hamming_weight < 0 or not isinstance(hamming_weight, int):
            raise ValueError(f"Invalid hamming weight value. Got {hamming_weight}.")

        if hamming_weight > n_qubits:
            raise ValueError(
                f"Hamming weight larger than number of qubits. \
                    Got {hamming_weight}. Max can be {n_qubits}."
            )

        if hamming_weight == 0:
            return initial_wf
        else:
            del initial_wf

            # Get first value with hamming weight
            current_value = int("1" * hamming_weight, base=2)

            counter: int = 1
            indices: List[int] = [current_value]
            while True:
                current_value = _get_next_number_with_same_hamming_weight(current_value)
                if not _most_significant_set_bit(current_value) <= n_qubits:
                    break
                indices.append(current_value)
                counter += 1

            amplitude = 1 / np.sqrt(counter)
            wf = np.zeros(2**n_qubits, dtype=np.complex128)
            wf[indices] = amplitude

            return Wavefunction(wf)

    def bind(self, symbol_map: Dict[Symbol, Any]) -> "Wavefunction":
        if not self.free_symbols:
            return self
        assert isinstance(self._amplitude_vector, Matrix)
        result = self._amplitude_vector.subs(symbol_map)

        try:
            return type(self)(result)
        except ValueError:
            raise ValueError("Passed map results in a violation of probability unity.")

    def get_probabilities(self) -> np.ndarray:
        return np.abs(self.amplitudes) ** 2

    def get_outcome_probs(self) -> Dict[str, float]:
        values = [
            format(i, "0" + str(self.n_qubits) + "b")[::-1] for i in range(len(self))
        ]

        probs = self.get_probabilities()

        return dict(zip(values, probs))


def flip_wavefunction(wavefunction: Wavefunction):
    return Wavefunction(flip_amplitudes(wavefunction.amplitudes))


def flip_amplitudes(amplitudes: Union[Sequence[complex], np.ndarray]) -> np.ndarray:
    number_of_states = len(amplitudes)
    ordering = _get_ordering(number_of_states)
    return np.asarray(amplitudes)[ordering]


@lru_cache
def _get_ordering(number_of_states: int) -> np.ndarray:
    num_bits = number_of_states.bit_length() - 1
    ordering = (
        np.arange(2**num_bits)
        .reshape(num_bits * [2])
        .transpose(*reversed(range(num_bits)))
        .reshape(2**num_bits)
    )
    return ordering


def load_wavefunction(file: LoadSource) -> Wavefunction:
    """Load a qubit wavefunction from a file.

    Args:
        file (str or file-like object): the name of the file, or a file-like object.

    Returns:
        wavefunction (orquestra.quantum.Wavefunction): the wavefunction object
    """

    with ensure_open(file) as f:
        data = json.load(f)

    wavefunction = Wavefunction(convert_dict_to_array(data["amplitudes"]))
    return wavefunction


def save_wavefunction(wavefunction: Wavefunction, filename: AnyPath) -> None:
    """Save a wavefunction object to a file.

    Args:
        wavefunction (orquestra.quantum.Wavefunction): the wavefunction object
        filename (str): the name of the file
    """

    data: Dict[str, Any] = {}
    data["amplitudes"] = convert_array_to_dict(wavefunction.amplitudes)
    with open(filename, "w") as f:
        f.write(json.dumps(data, indent=2))


def sample_from_wavefunction(
    wavefunction: Wavefunction,
    n_samples: int,
    seed: Optional[int] = None,
) -> List[Tuple[int, ...]]:
    """Sample bitstrings from a wavefunction.

    Args:
        wavefunction: the wavefunction to sample from.
        n_samples: the number of samples taken. Needs to be greater than 0.
        seed: the seed of the sampler

    Returns:
        List[Tuple[int]]: A list of tuples where the each tuple is a sampled bitstring.
    """
    if n_samples < 1:
        raise ValueError("Must sample from wavefunction at least once.")
    rng = np.random.default_rng(seed)
    outcome_strings, probabilities_np = zip(*wavefunction.get_outcome_probs().items())
    probabilities = [
        x[0] if isinstance(x, (list, np.ndarray)) else x for x in list(probabilities_np)
    ]
    # accelerate sampling by smartly choosing when formatting of samples is done
    if len(wavefunction) < n_samples:
        outcome_tuples: List[Union[Tuple[int, ...], int]] = []
        outcome_tuples += convert_bitstrings_to_tuples(outcome_strings)
        outcome_tuples += [0]  # adding non tuple forces rng.choice to return tuples
        probabilities += [0]  # need to add corresponding probability of 0
        samples = rng.choice(
            a=np.array(outcome_tuples, dtype=object), size=n_samples, p=probabilities
        ).tolist()
    else:
        string_samples = rng.choice(a=outcome_strings, size=n_samples, p=probabilities)
        samples = convert_bitstrings_to_tuples(string_samples)
    return samples



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
import operator
from functools import reduce

import numpy as np
import pytest

from orquestra.quantum.circuits import CNOT, RY, U3, Circuit, GateOperation, X, Y, Z
from orquestra.quantum.decompositions import (
    U3GateToRotation,
    decompose_orquestra_circuit,
)

DECOMPOSITION_RULES = [U3GateToRotation()]

U3_GATES = [
    U3(theta, phi, lambda_)
    for theta, phi, lambda_ in [
        (0, 0, 0),
        (0, np.pi, 0),
        (np.pi, 0, 0),
        (0, 0, np.pi),
        (np.pi / 2, np.pi / 2, np.pi / 2),
        (0.1 * np.pi, 0.5 * np.pi, 0.3 * np.pi),
        (4.1 * np.pi / 2, 2.5 * np.pi, 3 * np.pi),
    ]
]
CU3_GATES = [gate.controlled(1) for gate in U3_GATES]


def _is_scaled_identity(matrix: np.ndarray, rtol=1e-5, atol=1e-8) -> bool:
    """Returns whether 2 unitaries are equivalent up to a global phase.

    Usage:
    _is_scaled_identity(unitary_1 @ np.linalg.inv(unitary_2))
    """
    assert matrix.shape == (
        matrix.shape[0],
        matrix.shape[0],
    ), "This test is meaningful only for square matrices"

    target_matrix = np.diag(matrix).mean() * np.eye(
        matrix.shape[0], dtype=np.complex128
    )
    return np.allclose(matrix, target_matrix, rtol, atol)


class TestDecompositionOfU3Gates:
    @pytest.mark.parametrize(
        "gate_to_be_decomposed, target_qubits",
        [
            *[(gate, qubits) for gate in U3_GATES for qubits in [(0,), (2,)]],
        ],
    )
    def test_gives_the_same_unitary_as_original_gate_up_to_global_phase(
        self, gate_to_be_decomposed, target_qubits
    ):
        circuit = Circuit([gate_to_be_decomposed(*target_qubits)])
        decomposed_circuit = decompose_orquestra_circuit(circuit, DECOMPOSITION_RULES)

        assert _is_scaled_identity(
            circuit.to_unitary() @ np.linalg.inv(decomposed_circuit.to_unitary()),
        )

    @pytest.mark.parametrize(
        "operations",
        [[RY(np.pi / 2)(0)], [X(3), Y(1), Z(0)], [CNOT(3, 11)]],
    )
    def test_leaves_gates_not_matching_predicate_unaffected(self, operations):
        circuit = Circuit(operations)
        decomposed_circuit = decompose_orquestra_circuit(circuit, DECOMPOSITION_RULES)

        assert circuit.operations == decomposed_circuit.operations

    @pytest.mark.parametrize("target_qubits", [(0,), (2,)])
    @pytest.mark.parametrize("gate_to_be_decomposed", U3_GATES)
    def test_U3_decomposition_comprises_only_rotations(
        self, gate_to_be_decomposed, target_qubits
    ):
        circuit = Circuit([gate_to_be_decomposed(*target_qubits)])
        decomposed_circuit = decompose_orquestra_circuit(circuit, DECOMPOSITION_RULES)

        assert all(
            isinstance(op, GateOperation) and op.gate.name in ("RZ", "RY")
            for op in decomposed_circuit.operations
        )


class TestDecompositionOfCU3Gates:
    @pytest.mark.parametrize(
        "gate_to_be_decomposed, target_qubits",
        [
            *[(gate, qubits) for gate in CU3_GATES for qubits in [(0, 1)]],
        ],
    )
    def test_gives_the_same_unitary_as_original_gate_up_to_global_phase(
        self, gate_to_be_decomposed, target_qubits
    ):
        circuit = Circuit([gate_to_be_decomposed(*target_qubits)])
        decomposed_circuit = decompose_orquestra_circuit(circuit, DECOMPOSITION_RULES)

        def numpy_caster(x):
            return np.array(x).astype(np.complex128)

        original_matrix = numpy_caster(circuit.operations[0].gate.wrapped_gate.matrix)

        decomposed_matrix = reduce(
            operator.matmul,
            [
                op.gate.wrapped_gate.matrix
                for op in reversed(decomposed_circuit.operations)
            ],
        )

        decomposed_matrix = numpy_caster(decomposed_matrix)

        assert _is_scaled_identity(
            original_matrix @ np.linalg.inv(decomposed_matrix),
        )

    @pytest.mark.parametrize("target_qubits", [(0, 1), (3, 11)])
    @pytest.mark.parametrize("gate_to_be_decomposed", CU3_GATES)
    def test_CU3_decomposition_comprises_only_controlled_rotations(
        self, gate_to_be_decomposed, target_qubits
    ):
        circuit = Circuit([gate_to_be_decomposed(*target_qubits)])
        decomposed_circuit = decompose_orquestra_circuit(circuit, DECOMPOSITION_RULES)

        assert all(
            isinstance(op, GateOperation)
            and op.gate.name == "Control"
            and op.gate.wrapped_gate.name in ("RZ", "RY")
            for op in decomposed_circuit.operations
        )



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
import pytest
import sympy

from orquestra.quantum.api.circuit_runner_contracts import (
    CIRCUIT_RUNNER_CONTRACTS,
    STRICT_CIRCUIT_RUNNER_CONTRACTS,
    circuit_runner_gate_compatibility_contracts,
)
from orquestra.quantum.api.wavefunction_simulator_contracts import (
    simulator_contracts_for_tolerance,
    simulator_gate_compatibility_contracts,
)
from orquestra.quantum.circuits import U3, XX, Circuit
from orquestra.quantum.runners.symbolic_simulator import SymbolicSimulator


@pytest.fixture
def backend():
    return SymbolicSimulator()


@pytest.fixture
def wf_simulator() -> SymbolicSimulator:
    return SymbolicSimulator()


class TestSymbolicSimulator:
    @pytest.mark.parametrize(
        "gate",
        [
            XX(sympy.Symbol("theta"))(2, 1),
            U3(
                sympy.Symbol("alpha"),
                sympy.Symbol("beta"),
                sympy.Symbol("gamma"),
            )(1),
        ],
    )
    def test_cannot_sample_from_circuit_containing_free_symbols(self, gate):
        simulator = SymbolicSimulator()
        circuit = Circuit([gate])

        with pytest.raises(ValueError):
            simulator.run_and_measure(circuit, n_samples=1000)


@pytest.mark.parametrize("contract", CIRCUIT_RUNNER_CONTRACTS)
def test_symbolic_simulator_fullfills_circuit_runner_contracts(contract):
    simulator = SymbolicSimulator()
    assert contract(simulator)


@pytest.mark.parametrize("contract", simulator_contracts_for_tolerance())
def test_symbolic_simulator_fulfills_simulator_contracts(contract):
    simulator = SymbolicSimulator()
    assert contract(simulator)


@pytest.mark.parametrize("contract", STRICT_CIRCUIT_RUNNER_CONTRACTS)
def test_symbolic_simulator_fulfills_strict_circuit_runnner(contract):
    simulator = SymbolicSimulator()
    assert contract(simulator)


class TestSymbolicSimulatorUsesCorrectGateDefinitions:
    @pytest.mark.parametrize("contract", simulator_gate_compatibility_contracts())
    def test_using_target_amplitudes(self, contract):
        simulator = SymbolicSimulator()
        assert contract(simulator)

    @pytest.mark.parametrize("contract", circuit_runner_gate_compatibility_contracts())
    def test_using_expectation_values(self, contract):
        simulator = SymbolicSimulator(seed=1234)
        assert contract(simulator)



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
import math
from typing import TYPE_CHECKING, Dict

if TYPE_CHECKING:
    from ..distributions import MeasurementOutcomeDistribution


def compute_clipped_negative_log_likelihood(
    target_distribution: "MeasurementOutcomeDistribution",
    measured_distribution: "MeasurementOutcomeDistribution",
    distance_measure_parameters: Dict,
) -> float:
    """Compute the value of the clipped negative log likelihood between a target
    distribution and a measured distribution.

    See Equation (4) in https://advances.sciencemag.org/content/5/10/eaaw9918?rss=1

    Args:
        target_distribution: The target probability distribution.
        measured_distribution: The measured probability distribution.
        distance_measure_parameters:
            epsilon (float): The small parameter needed to regularize log computation
                when argument is zero. The default value is 1e-9.

    Returns:
        The value of the clipped negative log likelihood
    """

    epsilon = distance_measure_parameters.get("epsilon", 1e-9)
    value = 0.0
    target_keys = target_distribution.distribution_dict.keys()
    measured_keys = measured_distribution.distribution_dict.keys()
    all_keys = set(target_keys).union(measured_keys)

    for bitstring in all_keys:
        target_bitstring_value = target_distribution.distribution_dict.get(bitstring, 0)
        measured_bitstring_value = measured_distribution.distribution_dict.get(
            bitstring, 0
        )

        value += target_bitstring_value * math.log(
            max(epsilon, measured_bitstring_value)
        )

    return -value



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
import operator
from functools import reduce, singledispatch
from itertools import groupby
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

import numpy as np
import sympy

from . import _gates, _operations


def _circuit_size_by_operations(operations):
    return (
        0
        if not operations
        else max(
            qubit_index
            for operation in operations
            for qubit_index in operation.qubit_indices
        )
        + 1
    )


def _operation_uses_custom_gate(operation):
    return isinstance(operation.gate, _gates.MatrixFactoryGate) and isinstance(
        operation.gate.matrix_factory, _gates.CustomGateMatrixFactory
    )


class Circuit:
    """orquestra representation of a quantum circuit.

    See `help(orquestra.quantum.circuits)` for usage guide.
    """

    def __init__(
        self,
        operations: Optional[Iterable[_operations.Operation]] = None,
        n_qubits: Optional[int] = None,
    ):
        self._operations = list(operations) if operations is not None else []

        if n_qubits:
            cast_n_qubits = int(n_qubits)

            if n_qubits != cast_n_qubits:
                raise ValueError("Non-integer value passed.")

            if cast_n_qubits <= 0:
                raise ValueError("Non-positive value passed.")

            self._n_qubits = n_qubits
        else:
            self._n_qubits = _circuit_size_by_operations(self._operations)

    @property
    def operations(self):
        """Sequence of quantum gates to apply to qubits in this circuit."""
        return self._operations

    @property
    def n_qubits(self) -> int:
        """Number of qubits in this circuit.
        Not every qubit has to be used by a gate.
        """
        return self._n_qubits

    @property
    def free_symbols(self) -> List[sympy.Symbol]:
        """Set of all the sympy symbols used as params of gates in the circuit.
        The output list is sorted based on the order of appearance
        in `self._operations`."""
        seen_symbols = set()
        symbols_sequence = []
        for operation in self._operations:
            for symbol in operation.free_symbols:
                if symbol not in seen_symbols:
                    seen_symbols.add(symbol)
                    symbols_sequence.append(symbol)

        return symbols_sequence

    def __eq__(self, other: object):
        if not isinstance(other, type(self)):
            return NotImplemented

        if self.n_qubits != other.n_qubits:
            return False

        return list(self.operations) == list(other.operations)

    def __add__(self, other: Union["Circuit", _gates.GateOperation]):
        return _append_to_circuit(other, self)

    def collect_custom_gate_definitions(self) -> Iterable[_gates.CustomGateDefinition]:
        custom_gate_definitions = (
            operation.gate.matrix_factory.gate_definition
            for operation in self.operations
            if _operation_uses_custom_gate(operation)
        )
        unique_operation_dict = {}
        for gate_def in custom_gate_definitions:
            if gate_def.gate_name not in unique_operation_dict:
                unique_operation_dict[gate_def.gate_name] = gate_def
            elif unique_operation_dict[gate_def.gate_name] != gate_def:
                raise ValueError(
                    "Different gate definitions with the same name exist: "
                    f"{gate_def.gate_name}."
                )
        return sorted(
            unique_operation_dict.values(), key=operator.attrgetter("gate_name")
        )

    def to_unitary(self) -> Union[np.ndarray, sympy.Matrix]:
        """Create a unitary matrix describing Circuit's action.

        For performance reasons, this method will construct numpy matrix if circuit does
        not have free parameters, and a sympy matrix otherwise.
        """
        # The `reversed` iterator reflects the fact the matrices are multiplied
        # when composing linear operations (i.e. first operation is the rightmost).
        lifted_matrices = [
            op.lifted_matrix(self.n_qubits) for op in reversed(self.operations)
        ]
        return reduce(operator.matmul, lifted_matrices)

    def bind(self, symbols_map: Dict[sympy.Symbol, Any]):
        """Create a copy of the current circuit with the parameters of each gate bound
        to the values provided in the input symbols map.

        Args:
            symbols_map: A map of the symbols/gate parameters to new values
        """
        return type(self)(
            operations=[op.bind(symbols_map) for op in self.operations],
            n_qubits=self.n_qubits,
        )

    def __repr__(self):
        return (
            f"{type(self).__name__}"
            f"(operations=[{', '.join(map(str, self.operations))}], "
            f"n_qubits={self.n_qubits})"
        )

    def inverse(self) -> "Circuit":
        """Create a circuit that, when applied to the current circuit, will
        produce the identity circuit."""
        assert all(isinstance(op, _gates.GateOperation) for op in self.operations)
        try:
            return type(self)(
                operations=[
                    op.gate.dagger(*op.qubit_indices)
                    for op in reversed(self.operations)
                ],
                n_qubits=self.n_qubits,
            )
        except AttributeError as e:
            raise AttributeError(
                "Inverse is not implemented for this circuit type,"
                " since there are operators in it without the `dagger` method."
            ) from e


@singledispatch
def _append_to_circuit(other, circuit: Circuit):
    raise NotImplementedError()


@_append_to_circuit.register
def _append_operation(other: _gates.GateOperation, circuit: Circuit):
    n_qubits_by_operation = max(other.qubit_indices) + 1
    return type(circuit)(
        operations=[*circuit.operations, other],
        n_qubits=max(circuit.n_qubits, n_qubits_by_operation),
    )


@_append_to_circuit.register
def _append_circuit(other: Circuit, circuit: Circuit):
    return type(circuit)(
        operations=[*circuit.operations, *other.operations],
        n_qubits=max(circuit.n_qubits, other.n_qubits),
    )


def split_circuit(
    circuit: Circuit, predicate: Callable[[_operations.Operation], bool]
) -> Iterable[Tuple[bool, Circuit]]:
    """Split circuit into subcircuits for which predicate on all operation is constant.

    Args:
        circuit: a circuit to be split
        predicate: function assigning boolean value to each operation, its values
          are used for grouping operations belonging to the same subcircuits.
    Returns:
        An iterable of tuples of the form (x, subcircuit) s.t.:
        - predicate(operation) == x for every operation in subcircuit.operations
        - for two consecutive tuples (x1, subcircuit1), (x2, subcircuit2)
          x1 != x2 (i.e. consecutive chunks differ in the predicate value),
        - operations in subcircuits follow the same order as in original circuit
        - all subcircuits have the same number of qubits equal to `circuit.n_qubits`.
    """
    n_qubits = circuit.n_qubits
    for predicate_value, operations in groupby(circuit.operations, predicate):
        yield predicate_value, Circuit(operations, n_qubits=n_qubits)



################################################################################
# © Copyright 2020-2022 Zapata Computing Inc.
################################################################################
"""Functions for constructing circuits simulating evolution under given Hamiltonian."""
import operator
import warnings
from functools import reduce
from itertools import chain
from typing import List, Optional, Tuple, Union

import numpy as np
import sympy

from .circuits import CNOT, RX, RZ, Circuit, GateOperation, H
from .operators import PauliRepresentation, PauliTerm


def time_evolution(
    hamiltonian: PauliRepresentation,
    time: Union[float, sympy.Expr],
    method: str = "Trotter",
    trotter_order: int = 1,
) -> Circuit:
    """Create a circuit simulating evolution under given Hamiltonian.

    Args:
        hamiltonian: The Hamiltonian to be evolved under.
        time: Time duration of the evolution.
        method: Time evolution method. Currently the only option is 'Trotter'.
        trotter_order: order of Trotter evolution (1 by default).

    Returns:
        Circuit approximating evolution under `hamiltonian`.
        Circuit's unitary i approximately equal to exp(-i * time * hamiltonian).
    """
    if method != "Trotter":
        raise ValueError(f"Currently the method {method} is not supported.")

    return reduce(
        operator.add,
        (
            time_evolution_for_term(term, time / trotter_order)
            for _index_order in range(trotter_order)
            for term in hamiltonian.terms
        ),
    )


def time_evolution_for_term(term: PauliTerm, time: Union[float, sympy.Expr]) -> Circuit:
    """Evolves a Pauli term for a given time and returns a circuit representing it.
    Based on section 4 from https://arxiv.org/abs/1001.3855 .
    Args:
        term: Pauli term to be evolved
        time: time of evolution
    Returns:
        Circuit representing evolved term.
    """

    base_changes = []
    base_reversals = []
    cnot_gates = []
    central_gate: Optional[GateOperation] = None
    qubit_indices = sorted(term.qubits)

    circuit = Circuit()

    # If constant term, return empty circuit.
    if term.is_constant:
        return circuit

    if term.coefficient.imag > 1e-9:
        raise ValueError("Coefficients of terms must be real for Trotterization.")

    for i, qubit_id in enumerate(qubit_indices):
        term_type = term[qubit_id]
        if term_type == "X":
            base_changes.append(H(qubit_id))
            base_reversals.append(H(qubit_id))
        elif term_type == "Y":
            base_changes.append(RX(np.pi / 2)(qubit_id))
            base_reversals.append(RX(-np.pi / 2)(qubit_id))
        if i == len(term.operations) - 1:
            central_gate = RZ(2 * time * term.coefficient.real)(qubit_id)
        else:
            cnot_gates.append(CNOT(qubit_id, qubit_indices[i + 1]))

    for gate in base_changes:
        circuit += gate

    for gate in cnot_gates:
        circuit += gate

    if central_gate is not None:
        circuit += central_gate

    for gate in reversed(cnot_gates):
        circuit += gate

    for gate in base_reversals:
        circuit += gate

    return circuit


def time_evolution_derivatives(
    hamiltonian: PauliRepresentation,
    time: float,
    method: str = "Trotter",
    trotter_order: int = 1,
) -> Tuple[List[Circuit], List[float]]:
    """Generates derivative circuits for the time evolution operator defined in
    function time_evolution

    Args:
        hamiltonian: The Hamiltonian to be evolved under. It should contain numeric
            coefficients, symbolic expressions aren't supported.
        time: time duration of the evolution.
        method: time evolution method. Currently the only option is 'Trotter'.
        trotter_order: order of Trotter evolution

    Returns:
        A Circuit simulating time evolution.
    """
    if method != "Trotter":
        raise ValueError(f"The method {method} is currently not supported.")

    single_trotter_derivatives = []
    factors = [1.0, -1.0]
    output_factors = []
    terms = hamiltonian.terms

    for i, term_1 in enumerate(terms):
        for factor in factors:
            output = Circuit()

            if term_1.coefficient.real != term_1.coefficient:
                warnings.warn(
                    "Only real coefficients are supported. The imaginary part of the "
                    "term {} will be ignored.".format(term_1)
                )
            r = term_1.coefficient.real / trotter_order
            output_factors.append(r * factor)
            shift = factor * (np.pi / (4.0 * r))

            for j, term_2 in enumerate(terms):
                output += time_evolution_for_term(
                    term_2,
                    (time + shift) / trotter_order if i == j else time / trotter_order,
                )

            single_trotter_derivatives.append(output)

    if trotter_order > 1:
        output_circuits = []
        final_factors = []

        repeated_circuit = time_evolution(
            hamiltonian, time, method="Trotter", trotter_order=1
        )

        for position in range(trotter_order):
            for factor, different_circuit in zip(
                output_factors, single_trotter_derivatives
            ):
                output_circuits.append(
                    _generate_circuit_sequence(
                        repeated_circuit, different_circuit, trotter_order, position
                    )
                )
                final_factors.append(factor)
        return output_circuits, final_factors
    else:
        return single_trotter_derivatives, output_factors


def _generate_circuit_sequence(
    repeated_circuit: Circuit,
    different_circuit: Circuit,
    length: int,
    position: int,
):
    """Join multiple copies of circuit, replacing one copy with a different circuit.

    Args:
        repeated_circuit: circuit which copies should be concatenated
        different_circuit: circuit that will replace one copy of `repeated_circuit
        length: total number of circuits to join
        position: which copy of repeated_circuit should be replaced by
        `different_circuit`.
    Returns:
        Concatenation of circuits C_1, ..., C_length, where C_i = `repeated_circuit`
        if i != position and C_i = `different_circuit` if i == position.
    """
    if position >= length:
        raise ValueError(f"Position {position} should be < {length}")

    return Circuit(
        list(
            chain.from_iterable(
                [
                    (
                        repeated_circuit if i != position else different_circuit
                    ).operations
                    for i in range(length)
                ]
            )
        )
    )



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
from typing import TYPE_CHECKING, Dict

from .clipped_negative_log_likelihood import compute_clipped_negative_log_likelihood

if TYPE_CHECKING:
    from ..distributions import MeasurementOutcomeDistribution


def compute_jensen_shannon_divergence(
    target_distribution: "MeasurementOutcomeDistribution",
    measured_distribution: "MeasurementOutcomeDistribution",
    distance_measure_parameters: Dict,
) -> float:
    """Computes the symmetrized version of the clipped negative log likelihood between a
    target distribution and a measured distribution.
    See Equation (4) in https://advances.sciencemag.org/content/5/10/eaaw9918?rss=1

    Args:
        target_distribution: The target probability distribution.
        measured_distribution: The measured probability distribution.

        distance_measure_parameters:
            epsilon (float): The small parameter needed to regularize log computation
            when argument is zero. The default value is 1e-9.

    Returns:
        float: The value of the symmetrized version
    """

    value = (
        compute_clipped_negative_log_likelihood(
            target_distribution, measured_distribution, distance_measure_parameters
        )
        / 2
        + compute_clipped_negative_log_likelihood(
            measured_distribution, target_distribution, distance_measure_parameters
        )
        / 2
    )

    return value



################################################################################
# © Copyright 2020-2022 Zapata Computing Inc.
################################################################################
#   Copyright 2017 The OpenFermion Developers
#   Modifications copyright 2022 Zapata Computing, Inc. for compatibility reasons.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""This module contains functions from openfermion adapted for use with Orquestra's
PauliTerm and PauliSum classes.
"""
from .config import EQ_TOLERANCE
from .operator_utils import hermitian_conjugated, is_hermitian
from .sparse_tools import expectation, get_sparse_operator



from collections import Counter
from functools import reduce
from itertools import islice
from math import ceil
from typing import Dict, Iterable, List, Sequence, Tuple, TypeVar

T = TypeVar("T")


def _iterate_in_batches(items: Iterable[T], batch_size: int) -> Iterable[Tuple[T, ...]]:
    it = iter(items)
    while chunk := tuple(islice(it, batch_size)):
        yield chunk


def split_into_batches(
    circuits: Sequence[T],
    n_samples_per_circuit: Sequence[int],
    max_batch_size: int,
) -> Iterable[Tuple[Sequence[T], int]]:
    """Split sequence of circuits and corresponding sample sizes into batches.

    The circuits are split into batches of length `max_batch_size`, except
    possibly the last batch that can be smaller. The number of samples
    for each batch is a maximum number of samples requested for the circuits
    in the chunk. To illustrate, consider the following input:

    - circuits=[c1, c2, c3, c4, c5]
    - n_samples_per_circuit=[n1, n2, n3, n4, n5]
    - max_batch_size=2

    Then the returned iterable contains the following tuples:

    ((c1, c2), max([n1, n2]))
    ((c3, c4), max([n3, n4]))
    ((c5,), n5)

    Args:
        circuits: a sequence of circuits to be batched. Please note that
          the exact type of the circuits does not matter, and hence
          this function works for native Orquestra circuits as well as
          circuits from other libraries/SDKs.
        n_samples_per_circuit: a sequence of numbers of length equal to
          len(circuits), s.t. n_samples_per_circuit[i] is number of
          samples requested for circuits[i].
        max_batch_size: maximum allowable size of batch. Typically it comes
          from specific runner's restrictions.
    Returns:
        An iterable yielding tuples of the forms (circuits_chunk, n_samples)
        satisfying the following constraints:

        - len(circuits_chunk) <= max_batch_size
        - n_samples is large enough to accommodate sample sizes requested
          for each circuit in circuits_chunk

    Raises:
        ValueError: if input sequences are not of equal length, or if
          max_batch_size is not positive.
    """
    if len(circuits) != len(n_samples_per_circuit):
        raise ValueError(
            "Mismatched lengths of `circuits` and `n_samples_per_circuit: "
            f"({len(circuits)} and {len(n_samples_per_circuit)} respectively)."
            "Both sequences need to have the same length"
        )

    if max_batch_size <= 0:
        raise ValueError(
            "Max circuits per batch has to be positive, got " f"{max_batch_size}"
        )

    return (
        (circuits_chunk, max(samples_chunk))
        for circuits_chunk, samples_chunk in zip(
            _iterate_in_batches(circuits, max_batch_size),
            _iterate_in_batches(n_samples_per_circuit, max_batch_size),
        )
    )


def _expand_sample_size(n_samples, max_sample_size):
    multiplicities = ceil(n_samples / max_sample_size)
    new_n_samples = (
        multiplicities * (max_sample_size,)
        if n_samples % max_sample_size == 0
        else (multiplicities - 1) * (max_sample_size,) + (n_samples % max_sample_size,)
    )
    return new_n_samples, multiplicities


def expand_sample_sizes(
    circuits: Sequence[T], n_samples_per_circuit: Sequence[int], max_sample_size: int
) -> Tuple[Sequence[T], Sequence[int], Sequence[int]]:
    """Expand sample sizes for each circuit to fit maximum sample size.

    Args:
        circuits: list of circuits to be expanded
        n_samples_per_circuit: list of sample sizes corresponding to each
          circuit
        max_sample_size: maximum allowable sample size
    Returns:
      Tuple of three sequences (new_circuits, new_sample_sizes, multiplicities):

      - new_circuits: sequence of circuits containing all of the original, possibly
        duplicated, circuits
      - new_n_samples: list of integers of the same length as new_circuits,
        s.t. sum of all sample sizes corresponding to given circuit (counting
        duplicates) is equal to the original sample size
      - multiplicities: sequence of the same length as original sequence of
        circuits holding information on how many times given circuit has
        been duplicated in `new_circuits`
    """
    n_samples_and_multiplicities = [
        _expand_sample_size(n_samples, max_sample_size)
        for n_samples in n_samples_per_circuit
    ]

    new_n_samples = [
        n for n_samples, _ in n_samples_and_multiplicities for n in n_samples
    ]

    multiplicities = [multi for _, multi in n_samples_and_multiplicities]
    new_circuits = [
        circuit
        for circuit, multi in zip(circuits, multiplicities)
        for _ in range(multi)
    ]

    return new_circuits, new_n_samples, multiplicities


def _combine_measurements(
    first: Dict[str, int], second: Dict[str, int]
) -> Dict[str, int]:
    result = Counter(first)
    for bitstring, count in second.items():
        result[bitstring] += count
    return dict(result)


def combine_measurement_counts(
    all_measurements: Sequence[Dict[str, int]], multiplicities: Sequence[int]
) -> List[Dict[str, int]]:
    """Combine (aggregate) measurements of the same circuits run several times.

    Suppose multiplicities is a list [1, 2 ,3]. Then, the all_measurements should
    be a sequence of 1+2+3=6 elements m0,m1,m2,m3,m4,m5, and the results will
    contain three dictionaries M0, M1, M2 s.t.

    - M0 contains counts from m0 only
    - M1 comprises combined counts from m1 and m2
    - M2 comprises combined counts from m3, m4 and m5

    Args:
        all_measurements: sequence of measurements containing measurements
          gathered from some, possibly duplicated, circuits. The Measurement
          objects corresponding to the same circuit should be placed next to
          each other. Should have the same length as sum(multiplicities).
        multiplicities: sequence of positive integers marking groups of
          consecutive measurements corresponding to the same circuit. For
          instance, multiplicities [1, 2, 3] mean that first group of
          measurements comprises 1 Measurement, second group comprises 2
          consecutive Measurements, third group contains 3 consecutive
          Measurements and so on.
    Returns:
        Sequence of combined measurements of length equal len(multiplicities)
    Raises:
        ValueError: if len(all_measurements) != sum(multiplicities)
    """
    if len(all_measurements) != (sum_multiplicities := sum(multiplicities)):
        raise ValueError(
            "Mismatch between multiplicities and number of measurements to combine. "
            f"Got {len(all_measurements)} Measurements objects to combine "
            f"but multiplicities sum to {sum_multiplicities}"
        )
    measurements_it = iter(all_measurements)
    return [
        reduce(_combine_measurements, islice(measurements_it, multiplicity))
        for multiplicity in multiplicities
    ]


def combine_bitstrings(
    all_bitstrings: Sequence[List[str]], multiplicities: Sequence[int]
) -> List[List[str]]:
    """Combine (aggregate) bitstrings of the same circuits run several times.

    Suppose multiplicities is a list [1, 2 ,3]. Then, the all_bitstrings should
    be a sequence of 1+2+3=6 elements, each of them being a sequence of bitstrings.
    For instance, all_bitstrings could be equal to:
    [["00", "01"], ["1", "0"], ["0"], ["001", "001"], ["111", "000"], ["000"]]
    and then the result would be:
    [
        ["00", "01"],
        ["1", "0", "0],
        ["001", "001", "111", "000", "000"]
    ]

    Args:
        all_bitstrings: sequence of lists containing bitstrings
          gathered from some, possibly duplicated, circuits. The bitstrings
          lists objects corresponding to the same circuit should be placed next
          to each other. Should have the same length as sum(multiplicities).
        multiplicities: sequence of positive integers marking groups of
          consecutive measurements corresponding to the same circuit. For
          instance, multiplicities [1, 2, 3] mean that first group of
          bitstrings comprises 1 sequence, second group comprises 2
          consecutive sequences, third group contains 3 consecutive
          sequences and so on.
    Returns:
        Sequence of combined measurements of length equal len(multiplicities)
    Raises:
        ValueError: if len(all_bitstrings) != sum(multiplicities)
    """
    if len(all_bitstrings) != (sum_multiplicities := sum(multiplicities)):
        raise ValueError(
            "Mismatch between multiplicities and number of measurements to combine. "
            f"Got {len(all_bitstrings)} bitstrings lists objects to combine "
            f"but multiplicities sum to {sum_multiplicities}"
        )
    bitstrings_it = iter(all_bitstrings)
    return [
        sum(islice(bitstrings_it, multiplicity), start=[])
        for multiplicity in multiplicities
    ]



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
import math
import warnings
from unittest import mock

import pytest

from orquestra.quantum.distributions import (
    MeasurementOutcomeDistribution,
    compute_clipped_negative_log_likelihood,
    compute_jensen_shannon_divergence,
    compute_mmd,
    evaluate_distribution_distance,
)


def test_clipped_negative_log_likelihood_is_computed_correctly():
    target_distr = MeasurementOutcomeDistribution({"000": 0.5, "111": 0.5})
    measured_dist = MeasurementOutcomeDistribution({"000": 0.1, "111": 0.9})
    distance_measure_params = {"epsilon": 0.1}
    clipped_log_likelihood = compute_clipped_negative_log_likelihood(
        target_distr, measured_dist, distance_measure_params
    )

    assert clipped_log_likelihood == 1.203972804325936


def test_uses_epsilon_instead_of_zero_in_target_distribution():
    log_spy = mock.Mock(wraps=math.log)
    with mock.patch("orquestra.quantum.distributions.math.log", log_spy):
        target_distr = MeasurementOutcomeDistribution(
            {"000": 0.5, "111": 0.4, "010": 0.0}
        )
        measured_dist = MeasurementOutcomeDistribution(
            {"000": 0.1, "111": 0.9, "010": 0.0}
        )
        distance_measure_params = {"epsilon": 0.01}
        compute_clipped_negative_log_likelihood(
            target_distr, measured_dist, distance_measure_params
        )

        log_spy.assert_has_calls(
            [mock.call(0.1), mock.call(0.9), mock.call(0.01)], any_order=True
        )


@pytest.mark.parametrize(
    "measured_dist,distance_measure_params,expected_mmd",
    [
        (
            MeasurementOutcomeDistribution({"000": 0.1, "111": 0.9}),
            {"sigma": 0.5},
            0.32000000000000006,
        ),
        (
            MeasurementOutcomeDistribution({"000": 0.5, "111": 0.5}),
            {"sigma": 1},
            0.00,
        ),
        (
            MeasurementOutcomeDistribution({"000": 0.5, "111": 0.5}),
            {"sigma": [1, 0.5, 2]},
            0.00,
        ),
    ],
)
def test_gaussian_mmd_is_computed_correctly(
    measured_dist, distance_measure_params, expected_mmd
):
    """Maximum mean discrepancy (MMD) with gaussian kernel between distributions is
    computed correctly."""
    target_distr = MeasurementOutcomeDistribution({"000": 0.5, "111": 0.5})
    mmd = compute_mmd(target_distr, measured_dist, distance_measure_params)

    assert mmd == expected_mmd


@pytest.mark.parametrize(
    "distance_measure_function, expected_default_values",
    [
        (
            compute_mmd,
            {"sigma": 1.0},
        ),
        (compute_clipped_negative_log_likelihood, {"epsilon": 1e-9}),
        (compute_jensen_shannon_divergence, {"epsilon": 1e-9}),
    ],
)
def test_distance_measure_default_parameters_are_set_correctly(
    distance_measure_function, expected_default_values
):
    target_distr = MeasurementOutcomeDistribution({"000": 0.5, "111": 0.5})
    measured_distr = MeasurementOutcomeDistribution({"000": 0.1, "111": 0.9})
    distance = distance_measure_function(target_distr, measured_distr, {})
    expected_distance = distance_measure_function(
        target_distr, measured_distr, expected_default_values
    )

    assert distance == expected_distance


@pytest.mark.parametrize(
    "target_cls,measured_cls,distance_measure",
    [
        (MeasurementOutcomeDistribution, dict, compute_clipped_negative_log_likelihood),
        (dict, MeasurementOutcomeDistribution, compute_clipped_negative_log_likelihood),
        (dict, dict, compute_clipped_negative_log_likelihood),
        (MeasurementOutcomeDistribution, dict, compute_mmd),
        (dict, MeasurementOutcomeDistribution, compute_mmd),
        (dict, dict, compute_mmd),
        (MeasurementOutcomeDistribution, dict, compute_jensen_shannon_divergence),
        (dict, MeasurementOutcomeDistribution, compute_jensen_shannon_divergence),
        (dict, dict, compute_jensen_shannon_divergence),
    ],
)
def test_distribution_distance_can_be_evaluated_only_for_bitstring_distributions(
    target_cls, measured_cls, distance_measure
):
    target = target_cls({"0": 10, "1": 5})
    measured = measured_cls({"0": 10, "1": 5})

    with pytest.raises(TypeError):
        evaluate_distribution_distance(target, measured, distance_measure)


@pytest.mark.parametrize(
    "distance_measure",
    [
        compute_clipped_negative_log_likelihood,
        compute_mmd,
        compute_jensen_shannon_divergence,
    ],
)
def test_distribution_distance_cannot_be_evaluated_if_supports_are_incompatible(
    distance_measure,
):
    target = MeasurementOutcomeDistribution({"0": 10, "1": 5})
    measured = MeasurementOutcomeDistribution({"00": 10, "10": 5})

    with pytest.raises(RuntimeError):
        evaluate_distribution_distance(target, measured, distance_measure)


@pytest.mark.parametrize(
    "normalize_target,normalize_measured, distance_measure",
    [
        (True, False, compute_clipped_negative_log_likelihood),
        (False, True, compute_clipped_negative_log_likelihood),
        (True, False, compute_mmd),
        (False, True, compute_mmd),
        (True, False, compute_jensen_shannon_divergence),
        (False, True, compute_jensen_shannon_divergence),
    ],
)
def test_distribution_distance_cant_be_computed_if_only_one_distribution_is_normalized(
    normalize_target, normalize_measured, distance_measure
):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", UserWarning)
        target = MeasurementOutcomeDistribution({"0": 10, "1": 5}, normalize_target)
        measured = MeasurementOutcomeDistribution({"0": 10, "1": 5}, normalize_measured)

    with pytest.raises(RuntimeError):
        evaluate_distribution_distance(target, measured, distance_measure)


def test_jensen_shannon_divergence_is_computed_correctly():
    """jensen shannon divergence between distributions is computed correctly."""
    target_distr = MeasurementOutcomeDistribution({"000": 0.5, "111": 0.5})
    measured_dist = MeasurementOutcomeDistribution({"000": 0.1, "111": 0.9})
    distance_measure_params = {"epsilon": 0.1}
    jensen_shannon_divergence = compute_jensen_shannon_divergence(
        target_distr, measured_dist, distance_measure_params
    )

    assert jensen_shannon_divergence == 0.9485599924429406



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
"""Utilities for converting sympy expressions to our native Expression format."""
import operator
from functools import singledispatch
from numbers import Number

import sympy
import sympy.core

try:
    # for sympy>=1.6 - import the module as usual
    import sympy.core.numbers as sympy_numbers
except ImportError:
    # for sympy<=1.5.1
    # There's a wildcard import in sympy that messes up the imports. This is a
    # workaround to be able to work with sympy numbers from both old and recent sympy
    # versions. See more at:
    # https://github.com/sympy/sympy/blob/70381f282f2d9d039da860e391fe51649df2779d/sympy/__init__.py#L57
    sympy_numbers = sympy.numbers  # type: ignore


from .expressions import ExpressionDialect, FunctionCall, Symbol, reduction


def is_multiplication_by_reciprocal(sympy_mul: sympy.Mul) -> bool:
    """Check if given sympy multiplication is of the form x * (1 / y)."""
    args = sympy_mul.args
    return len(args) == 2 and isinstance(args[1], sympy.Pow) and args[1].args[1] == -1


def is_addition_of_negation(sympy_add: sympy.Add) -> bool:
    """Check if given sympy addition is of the form x + (-y)."""
    args = sympy_add.args
    return len(args) == 2 and isinstance(args[1], sympy.Mul) and args[1].args[0] == -1


@singledispatch
def expression_from_sympy(expression):
    """Parse Sympy expression into intermediate expression tree."""
    raise NotImplementedError(
        f"Expression {expression} of type {type(expression)} is currently not supported"
    )


@expression_from_sympy.register
def identity(number: Number):
    return number


@expression_from_sympy.register
def symbol_from_sympy(symbol: sympy.Symbol):
    return Symbol(str(symbol))


@expression_from_sympy.register
def native_integer_from_sympy_integer(number: sympy.Integer):
    return int(number)


@expression_from_sympy.register
def native_float_from_sympy_float(number: sympy.Float):
    return float(number)


@expression_from_sympy.register
def native_float_from_sympy_rational(number: sympy.Rational):
    return float(number)


@expression_from_sympy.register
def native_imaginary_unit_from_sympy_imaginary_unit(
    _unit: sympy_numbers.ImaginaryUnit,
):
    return 1j


def _negate_sympy_expr(expr):
    return expr * (-1)


@expression_from_sympy.register
def addition_from_sympy_add(add: sympy.Add):
    if is_addition_of_negation(add):
        return FunctionCall(
            "sub",
            (
                expression_from_sympy(add.args[0]),
                expression_from_sympy(_negate_sympy_expr(add.args[1])),
            ),
        )
    return FunctionCall("add", expression_from_sympy(add.args))


@expression_from_sympy.register
def multiplication_from_sympy_mul(mul: sympy.Mul):
    if is_multiplication_by_reciprocal(mul):
        return FunctionCall(
            "div",
            (
                expression_from_sympy(mul.args[0]),
                expression_from_sympy(mul.args[1].args[0]),
            ),
        )
    else:
        return FunctionCall("mul", expression_from_sympy(mul.args))


@expression_from_sympy.register
def power_from_sympy_pow(power: sympy.Pow):
    if power.args[1] == -1:
        return FunctionCall("div", (1, expression_from_sympy(power.args[0])))
    elif power.args[1] == 0.5:
        return FunctionCall("sqrt", (expression_from_sympy(power.args[0]),))
    else:
        return FunctionCall("pow", expression_from_sympy(power.args))


@expression_from_sympy.register
def function_call_from_sympy_function(function: sympy.Function):
    return FunctionCall(str(function.func), expression_from_sympy(function.args))


@expression_from_sympy.register
def expression_tuple_from_tuple_of_sympy_args(args: tuple):
    return tuple(expression_from_sympy(arg) for arg in args)


# Dialect defining conversion of intermediate expression tree to
# the expression based on Sympy functions/Symbols
# This is intended to be passed by a `dialect` argument of `translate_expression`.
SYMPY_DIALECT = ExpressionDialect(
    symbol_factory=lambda symbol: sympy.Symbol(symbol.name),
    number_factory=lambda number: number,
    known_functions={
        "add": reduction(operator.add),
        "mul": reduction(operator.mul),
        "div": operator.truediv,
        "sub": operator.sub,
        "pow": operator.pow,
        "cos": sympy.cos,
        "sin": sympy.sin,
        "exp": sympy.exp,
        "sqrt": sympy.sqrt,
        "tan": sympy.tan,
    },
)



#   Copyright 2017 The OpenFermion Developers
#   Modifications copyright 2022 Zapata Computing, Inc. for compatibility reasons.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""This module provides functions to interface with scipy.sparse."""
from functools import reduce
from typing import List, Optional

import numpy
import scipy
import scipy.sparse
import scipy.sparse.linalg

from .._pauli_operators import PauliRepresentation, PauliSum, PauliTerm

# Make global definitions.
identity_csc = scipy.sparse.identity(2, format="csc", dtype=complex)
pauli_x_csc = scipy.sparse.csc_matrix([[0.0, 1.0], [1.0, 0.0]], dtype=complex)
pauli_y_csc = scipy.sparse.csc_matrix([[0.0, -1.0j], [1.0j, 0.0]], dtype=complex)
pauli_z_csc = scipy.sparse.csc_matrix([[1.0, 0.0], [0.0, -1.0]], dtype=complex)
pauli_matrix_map = {
    "I": identity_csc,
    "X": pauli_x_csc,
    "Y": pauli_y_csc,
    "Z": pauli_z_csc,
}


def get_sparse_operator(operator: PauliRepresentation, n_qubits: Optional[int] = None):
    """Initialize a Scipy sparse matrix from a PauliTerm or PauliSum.

    Args:
        operator: the pauli representation to convert to matrix representation.
        n_qubits (int): Number of qubits.

    Returns:
        The corresponding Scipy sparse matrix.
    """
    if not isinstance(operator, (PauliSum, PauliTerm)):
        raise TypeError(
            "Failed to convert a {} to a sparse matrix.".format(type(operator).__name__)
        )

    if n_qubits is None:
        n_qubits = operator.n_qubits
    if n_qubits < operator.n_qubits:
        raise ValueError("Invalid number of qubits specified.")

    # Construct the Scipy sparse matrix.
    n_hilbert = 2**n_qubits
    values_list: List[numpy.ndarray] = []
    row_list: List[numpy.ndarray] = []
    column_list: List[numpy.ndarray] = []

    # Loop through the terms.
    for qubit_term in operator.terms:
        tensor_factor = 0
        coefficient = qubit_term.coefficient
        sparse_operators = [coefficient]
        for qubit_num, operator_str in sorted(qubit_term.operations):

            # Grow space for missing identity operators.
            if qubit_num > tensor_factor:
                identity_qubits = qubit_num - tensor_factor
                identity = scipy.sparse.identity(
                    2**identity_qubits, dtype=complex, format="csc"
                )
                sparse_operators += [identity]

            # Add actual operator to the list.
            sparse_operators += [pauli_matrix_map[operator_str]]
            tensor_factor = qubit_num + 1

        # Grow space at end of string unless operator acted on final qubit.
        if tensor_factor < n_qubits or not qubit_term:
            identity_qubits = n_qubits - tensor_factor
            identity = scipy.sparse.identity(
                2**identity_qubits, dtype=complex, format="csc"
            )
            sparse_operators += [identity]

        # Extract triplets from sparse_term.
        sparse_matrix = _kronecker_operators(sparse_operators)
        values_list.append(sparse_matrix.tocoo(copy=False).data)
        (column, row) = sparse_matrix.nonzero()
        column_list.append(column)
        row_list.append(row)

    # Create sparse operator.
    values_list = numpy.concatenate(values_list)
    row_list = numpy.concatenate(row_list)
    column_list = numpy.concatenate(column_list)
    sparse_operator = scipy.sparse.coo_matrix(
        (values_list, (row_list, column_list)), shape=(n_hilbert, n_hilbert)
    ).tocsc(copy=False)
    sparse_operator.eliminate_zeros()
    return sparse_operator


def _kronecker_operators(*args):
    """Return the Kronecker product of multiple sparse.csc_matrix operators."""
    return reduce(_wrapped_kronecker, *args)


def _wrapped_kronecker(operator_1, operator_2):
    """Return the Kronecker product of two sparse.csc_matrix operators."""
    return scipy.sparse.kron(operator_1, operator_2, "csc")


def expectation(operator, state):
    """Compute the expectation value of an operator with a state.

    Args:
        operator(scipy.sparse.spmatrix or scipy.sparse.linalg.LinearOperator):
            The operator whose expectation value is desired.
        state(numpy.ndarray or scipy.sparse.spmatrix): A numpy array
            representing a pure state or a sparse matrix representing a density
            matrix. If `operator` is a LinearOperator, then this must be a
            numpy array.

    Returns:
        A complex number giving the expectation value.

    Raises:
        ValueError: Input state has invalid format.
    """

    if isinstance(state, scipy.sparse.spmatrix):
        # Handle density matrix.
        if isinstance(operator, scipy.sparse.linalg.LinearOperator):
            raise ValueError(
                "Taking the expectation of a LinearOperator with "
                "a density matrix is not supported."
            )
        product = state * operator
        expectation = numpy.sum(product.diagonal())

    elif isinstance(state, numpy.ndarray):
        # Handle state vector.
        if len(state.shape) == 1:
            # Row vector
            expectation = numpy.dot(numpy.conjugate(state), operator * state)
        else:
            # Column vector
            expectation = numpy.dot(numpy.conjugate(state.T), operator * state)[0, 0]

    else:
        # Handle exception.
        raise ValueError("Input state must be a numpy array or a sparse matrix.")

    # Return.
    return expectation



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
"""Data structures for orquestra gates."""
import math
from dataclasses import dataclass, replace
from typing import Callable, Dict, Iterable, Protocol, Tuple, Union, runtime_checkable

import numpy as np
import sympy

from ..typing import ParameterizedVector
from ._operations import Parameter, get_free_symbols, sub_symbols
from ._unitary_tools import _lift_matrix_numpy, _lift_matrix_sympy

DAGGER_GATE_NAME = "Dagger"
CONTROLLED_GATE_NAME = "Control"
EXPONENTIAL_GATE_NAME = "Exponential"
POWER_GATE_SYMBOL = "^"


@runtime_checkable
class Gate(Protocol):
    """Interface of a quantum gate representable by a matrix, translatable to other
    frameworks and backends.

    See `orquestra.quantum.circuits` for a list of built-in gates and usage guide.
    """

    @property
    def name(self) -> str:
        """Globally unique name of the gate.

        Name is used in textual representation and dispatching in conversion between
        frameworks. Defining different gates with the same name as built-in ones
        is discouraged."""
        raise NotImplementedError()

    @property
    def params(self) -> Tuple[Parameter, ...]:
        """Value of parameters bound to this gate.

        Length of `params` should be equal to number of parameters in gate initializer.
        In particular, nonparametric gates should always return ().

        Examples:

        - an `H` gate has no params
        - a `RX(np.pi)` gate has a single param with value of `np.pi`
        - a `RX(sympy.Symbol("theta"))` gate has a single symbolic param `theta`
        - a `RX(sympy.sympify("theta * alpha"))` gate has a single symbolic expression
            param `theta*alpha`

        We need it for translations to other frameworks and for serialization.
        """
        raise NotImplementedError()

    @property
    def free_symbols(self) -> Iterable[sympy.Symbol]:
        """Unbound symbols in the gate matrix.

        Examples:

        - an `H` gate has no free symbols
        - a `RX(np.pi)` gate has no free symbols
        - a `RX(sympy.Symbol("theta"))` gate has a single free symbol `theta`
        - a `RX(sympy.sympify("theta * alpha"))` gate has two free symbols, `alpha` and
            `theta`
        - a `RX(sympy.sympify("theta * alpha")).bind({sympy.Symbol("theta"): 0.42})`
            gate has one free symbol, `alpha`
        """
        return get_free_symbols(self.params)

    @property
    def num_qubits(self) -> int:
        """Number of qubits this gate acts on.

        We need it because matrix is computed lazily, and we don't want to create matrix
        just to know the number of qubits.
        """
        raise NotImplementedError()

    @property
    def matrix(self) -> sympy.Matrix:
        """Unitary matrix describing gate's action on state vector.

        We need it to be able to implement .propagate() on the operation class.
        """
        raise NotImplementedError()

    def controlled(self, num_control_qubits: int) -> "Gate":
        raise NotImplementedError()

    @property
    def dagger(self) -> "Gate":
        raise NotImplementedError()

    def power(self, exponent: float) -> "Gate":
        """Gate representing the underlying matrix raised to the power
        of the given exponent.
        """
        raise NotImplementedError()

    @property
    def exp(self) -> "Gate":
        """Gate representing the exponential of the given gate."""
        raise NotImplementedError()

    def bind(self, symbols_map: Dict[sympy.Symbol, Parameter]) -> "Gate":
        raise NotImplementedError()

    def replace_params(self, new_params: Tuple[Parameter, ...]) -> "Gate":
        raise NotImplementedError()

    def __call__(self, *qubit_indices: int) -> "GateOperation":
        """Returns representation of applying this gate on qubits in a circuit."""
        return GateOperation(self, qubit_indices)


def gate_is_parametric(gate_ref, gate_params):
    return not not gate_params


@dataclass(frozen=True)
class GateOperation:
    """Represents applying a `Gate` to 1 or more qubits in a circuit."""

    gate: Gate
    qubit_indices: Tuple[int, ...]

    @property
    def params(self) -> Tuple[Parameter, ...]:
        return self.gate.params

    def bind(self, symbols_map: Dict[sympy.Symbol, Parameter]) -> "GateOperation":
        return GateOperation(self.gate.bind(symbols_map), self.qubit_indices)

    def replace_params(self, new_params: Tuple[Parameter, ...]) -> "GateOperation":
        return GateOperation(self.gate.replace_params(new_params), self.qubit_indices)

    def lifted_matrix(self, num_qubits):
        return (
            _lift_matrix_sympy(self.gate.matrix, self.qubit_indices, num_qubits)
            if self.gate.free_symbols
            else _lift_matrix_numpy(self.gate.matrix, self.qubit_indices, num_qubits)
        )

    def apply(self, amplitude_vector: ParameterizedVector) -> ParameterizedVector:
        num_qubits = np.log2(len(amplitude_vector))
        if 2**num_qubits != len(amplitude_vector):
            raise ValueError(
                "GateOperation can only be applied to multi-qubit state vector but "
                f"vector of length {len(amplitude_vector)} was provided."
            )

        return self.lifted_matrix(int(num_qubits)) @ amplitude_vector

    @property
    def free_symbols(self) -> Iterable[sympy.Symbol]:
        return self.gate.free_symbols

    def __str__(self):
        return f"{self.gate}({','.join(map(str, self.qubit_indices))})"


def _all_attrs_equal(obj, other_obj, attrs):
    return all(getattr(obj, attr) == getattr(other_obj, attr) for attr in attrs)


@dataclass(frozen=True)
class MatrixFactoryGate:
    """Data structure for a `Gate` with deferred matrix construction.

    Most built-in gates are instances of this class.
    See `orquestra.quantum.circuits` for built-in gates and usage guide.

    This class requires the gate definition to be present during deserialization, so
    it's not easily applicable for gates defined in Orquestra steps. If you want to
    define a new gate, check out `CustomGateDefinition` first.

    Keeping a `matrix_factory` instead of a plain gate matrix allows us to defer matrix
    construction to _after_ parameter binding. This saves unnecessary work in scenarios
    where we construct a quantum circuit and immediately bind parameter values. When
    done multiple times, e.g. for every gate in each optimization step, this can lead
    to major performance issues.

    Args:
        name: Name of this gate. Implementers of new gates should make sure that the
            names are unique.
        matrix_factory: a callable mapping arbitrary number of parameters into gate
            matrix. Implementers of new gates should make sure the returned matrices are
            square and of dimension being 2 ** `num_qubits`.
        params: gate parameters - either concrete values or opaque symbols.
            Will be passed to `matrix_factory` when `matrix` property is requested.
        num_qubits: number of qubits this gate acts on.
    """

    name: str
    matrix_factory: Callable[..., sympy.Matrix]
    params: Tuple[Parameter, ...]
    num_qubits: int
    is_hermitian: bool = False

    @property
    def matrix(self) -> sympy.Matrix:
        """Unitary matrix defining action of this gate.

        This is a computed property using `self.matrix_factory` called with parameters
        bound to this gate.
        """
        return self.matrix_factory(*self.params)

    def bind(self, symbols_map) -> "MatrixFactoryGate":
        return self.replace_params(
            tuple(sub_symbols(param, symbols_map) for param in self.params)
        )

    def replace_params(self, new_params: Tuple[Parameter, ...]) -> "MatrixFactoryGate":
        return replace(self, params=new_params)

    def controlled(self, num_controlled_qubits: int) -> Gate:
        return ControlledGate(self, num_controlled_qubits)

    @property
    def dagger(self) -> Union["MatrixFactoryGate", Gate]:
        return self if self.is_hermitian else Dagger(self)

    @property
    def exp(self) -> "Gate":
        return Exponential(self)

    def power(self, exponent: float) -> "Gate":
        return Power(self, exponent)

    def __str__(self):
        return (
            f"{self.name}({', '.join(map(str,self.params))})"
            if self.params
            else self.name
        )

    def __eq__(self, other):
        if type(self) != type(other):
            return False

        if not _all_attrs_equal(
            self, other, set(self.__dataclass_fields__) - {"params"}
        ):
            return False

        if len(self.params) != len(other.params):
            return False

        return all(
            _are_matrix_elements_equal(p1, p2)
            for p1, p2 in zip(self.params, other.params)
        )

    # Normally, we'd use the default implementations by inheriting from the Gate
    # protocol.  We can't do that because of __init__ arg default value issues, this is
    # the workaround.
    @property
    def free_symbols(self) -> Iterable[sympy.Symbol]:
        """Unbound symbols in the gate matrix. See Gate.free_symbols for details."""
        return get_free_symbols(self.params)

    __call__ = Gate.__call__


@dataclass(frozen=True)
class ControlledGate(Gate):
    wrapped_gate: Gate
    num_control_qubits: int

    def __post_init__(self):
        if self.num_control_qubits < 1:
            raise ValueError(
                f"Invalid number of control qubits. Got {self.num_control_qubits}"
            )

    @property
    def name(self):
        return CONTROLLED_GATE_NAME

    @property
    def num_qubits(self):
        return self.wrapped_gate.num_qubits + self.num_control_qubits

    @property
    def matrix(self):
        return sympy.Matrix.diag(
            sympy.eye(2**self.num_qubits - 2**self.wrapped_gate.num_qubits),
            self.wrapped_gate.matrix,
        )

    @property
    def params(self):
        return self.wrapped_gate.params

    def controlled(self, num_control_qubits: int) -> "ControlledGate":
        return ControlledGate(
            wrapped_gate=self.wrapped_gate,
            num_control_qubits=self.num_control_qubits + num_control_qubits,
        )

    @property
    def dagger(self) -> "ControlledGate":
        return ControlledGate(
            wrapped_gate=self.wrapped_gate.dagger,
            num_control_qubits=self.num_control_qubits,
        )

    @property
    def exp(self) -> "Gate":
        return Exponential(self)

    def power(self, exponent: float) -> "Gate":
        return ControlledGate(
            wrapped_gate=self.wrapped_gate.power(exponent),
            num_control_qubits=self.num_control_qubits,
        )

    def bind(self, symbols_map) -> "Gate":
        return self.wrapped_gate.bind(symbols_map).controlled(self.num_control_qubits)

    def replace_params(self, new_params: Tuple[Parameter, ...]) -> "Gate":
        return self.wrapped_gate.replace_params(new_params).controlled(
            self.num_control_qubits
        )

    def __str__(self):
        return self.num_control_qubits * "c-" + str(self.wrapped_gate)


@dataclass(frozen=True)
class Dagger(Gate):
    wrapped_gate: Gate

    @property
    def matrix(self) -> sympy.Matrix:
        return self.wrapped_gate.matrix.adjoint()

    @property
    def params(self) -> Tuple[Parameter, ...]:
        return self.wrapped_gate.params

    @property
    def num_qubits(self) -> int:
        return self.wrapped_gate.num_qubits

    @property
    def name(self):
        return self.wrapped_gate.name + "_" + DAGGER_GATE_NAME

    def controlled(self, num_control_qubits: int) -> Gate:
        return self.wrapped_gate.controlled(num_control_qubits).dagger

    def bind(self, symbols_map) -> "Gate":
        return self.wrapped_gate.bind(symbols_map).dagger

    def replace_params(self, new_params: Tuple[Parameter, ...]) -> "Gate":
        return self.wrapped_gate.replace_params(new_params).dagger

    @property
    def dagger(self) -> "Gate":
        return self.wrapped_gate

    @property
    def exp(self) -> "Gate":
        return Exponential(self)

    def power(self, exponent: float) -> "Gate":
        return Power(self, exponent)

    def __str__(self):
        wrapped_string = str(self.wrapped_gate)
        before_and_after_params = wrapped_string.split("(")
        before_and_after_params[0] += "†"
        return "(".join(before_and_after_params)


@dataclass(frozen=True)
class Exponential(Gate):
    wrapped_gate: Gate

    def __post_init__(self):
        if len(self.wrapped_gate.free_symbols) > 0:
            raise ValueError("Cannot be perform exponential on gates with free symbols")

    @property
    def matrix(self) -> sympy.Matrix:
        return self.wrapped_gate.matrix.exp()

    @property
    def params(self) -> Tuple[Parameter, ...]:
        return self.wrapped_gate.params

    @property
    def num_qubits(self) -> int:
        return self.wrapped_gate.num_qubits

    @property
    def name(self):
        return EXPONENTIAL_GATE_NAME

    def controlled(self, num_control_qubits: int) -> Gate:
        return ControlledGate(self, num_control_qubits)

    def bind(self, symbols_map) -> "Gate":
        raise NotImplementedError(
            "Gates exponential do not possess free symbols to bind"
        )

    def replace_params(self, new_params: Tuple[Parameter, ...]) -> "Gate":
        return self.wrapped_gate.replace_params(new_params).exp

    @property
    def dagger(self) -> "Gate":
        return self.wrapped_gate.dagger.exp

    @property
    def exp(self) -> "Gate":
        return Exponential(self)

    def power(self, exponent: float) -> "Gate":
        return Power(self, exponent)

    def __str__(self):
        if check_has_non_commuting_gate_type(self, [ControlledGate, Power]):
            return "exp" + POWER_GATE_SYMBOL + "{" + str(self.wrapped_gate) + "}"
        return "exp" + POWER_GATE_SYMBOL + str(self.wrapped_gate)


@dataclass(frozen=True)
class Power(Gate):
    wrapped_gate: Gate
    exponent: float

    def __post_init__(self):
        if len(self.wrapped_gate.free_symbols) > 0:
            raise ValueError("Cannot return power of gates with free symbols")

    @property
    def name(self) -> str:
        return f"{self.wrapped_gate.name}{POWER_GATE_SYMBOL}{self.exponent}"

    @property
    def params(self) -> Tuple[Parameter, ...]:
        return self.wrapped_gate.params

    @property
    def free_symbols(self) -> Iterable[sympy.Symbol]:
        return get_free_symbols(self.params)

    @property
    def num_qubits(self) -> int:
        return self.wrapped_gate.num_qubits

    @property
    def matrix(self) -> sympy.Matrix:
        return self.wrapped_gate.matrix**self.exponent

    def controlled(self, num_control_qubits: int) -> "Gate":
        return self.wrapped_gate.controlled(num_control_qubits).power(self.exponent)

    @property
    def dagger(self) -> "Gate":
        return self.wrapped_gate.dagger.power(self.exponent)

    @property
    def exp(self) -> "Gate":
        return Exponential(self)

    def power(self, exponent: float) -> "Gate":
        return Power(self, exponent)

    def bind(self, symbols_map: Dict[sympy.Symbol, Parameter]) -> "Gate":
        raise NotImplementedError(
            "Gates raised to a power do not possess free symbols to bind",
        )

    def replace_params(self, new_params: Tuple[Parameter, ...]) -> "Gate":
        return self.wrapped_gate.replace_params(new_params).power(self.exponent)

    def __str__(self):
        if check_has_non_commuting_gate_type(self, [Exponential]):
            inner_string = "{" + str(self.wrapped_gate) + "}"
        else:
            inner_string = str(self.wrapped_gate)

        return inner_string + POWER_GATE_SYMBOL + str(self.exponent)


def _n_qubits(matrix):
    n_qubits = math.floor(math.log2(matrix.shape[0]))
    if 2**n_qubits != matrix.shape[0] or 2**n_qubits != matrix.shape[1]:
        raise ValueError("Gate's matrix has to be square with dimension 2^N")
    return n_qubits


@dataclass(frozen=True)
class CustomGateMatrixFactory:
    """Can be passed as `matrix_factory` when a gate matrix isn't lazily evaluated."""

    gate_definition: "CustomGateDefinition"

    @property
    def matrix(self) -> sympy.Matrix:
        return self.gate_definition.matrix

    @property
    def params_ordering(self) -> Tuple[Parameter, ...]:
        return self.gate_definition.params_ordering

    def __call__(self, *gate_params):
        return self.matrix.subs(
            {symbol: arg for symbol, arg in zip(self.params_ordering, gate_params)}
        )

    def __eq__(self, other):
        if type(self) != type(other):
            return False

        if self.params_ordering != other.params_ordering:
            return False

        if not _are_matrices_equal(self.matrix, other.matrix):
            return False

        return True


@dataclass(frozen=True)
class CustomGateDefinition:
    """Use this class to define a non-built-in gate.

    See "Defining new gates" section in `help(orquestra.quantum.circuits)` for
    usage guide.

    User-defined gates are treated differently than the built-in ones,
    because the built-in ones are defined in `orquestra.quantum` library, and so
    we can assume that the definitions will be available during circuit deserialization.

    User-provided gates can be defined in one repo (e.g. Orquestra step), serialized,
    and passed to another project for deserialization. The other project must have
    access to gate details, e.g. the gate matrix. This class is designed to keep track
    of the gate details needed for deserialization.

    Instances of this class are serialized by the Circuit objects, additionally to
    Circuit operations.
    """

    gate_name: str
    matrix: sympy.Matrix
    params_ordering: Tuple[sympy.Symbol, ...]

    def __post_init__(self):
        n_qubits = _n_qubits(self.matrix)
        object.__setattr__(self, "_n_qubits", n_qubits)

    def __call__(self, *gate_params):
        return MatrixFactoryGate(
            self.gate_name,
            CustomGateMatrixFactory(self),
            gate_params,
            self._n_qubits,
        )

    def __eq__(self, other):
        if type(self) != type(other):
            return False

        if self.gate_name != other.gate_name:
            return False

        if self.params_ordering != other.params_ordering:
            return False

        if not _are_matrices_equal(self.matrix, other.matrix):
            return False

        return True


def _are_matrix_elements_equal(element, another_element):
    """Determine if two elements from gates' matrices are equal.

    This is to be used in __eq__ method when comparing matrices elementwise.

    Args:
        element: first value to compare. It can be float, complex or a sympy expression.
        another_element: second value to compare.
    """
    difference = sympy.N(sympy.expand(element) - sympy.expand(another_element))

    try:
        return np.allclose(
            float(sympy.re(difference)) + 1j * float(sympy.im(difference)), 0
        )
    except TypeError:
        return False


def _are_matrices_equal(matrix, another_matrix):
    return all(
        _are_matrix_elements_equal(element, another_element)
        for element, another_element in zip(matrix, another_matrix)
    )


def check_has_non_commuting_gate_type(self, non_commuting_gate_types):
    wrapped_gate = self.wrapped_gate
    has_non_commuting_gate_type = False
    while True:
        if type(self.wrapped_gate) in non_commuting_gate_types:
            has_non_commuting_gate_type = True
        if hasattr(wrapped_gate, "wrapped_gate"):
            wrapped_gate = wrapped_gate.wrapped_gate
        else:
            break
    return has_non_commuting_gate_type



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
import numpy as np
import pytest

from orquestra.quantum.circuits import (
    RY,
    U3,
    Circuit,
    X,
    Y,
    apply_gate_to_qubits,
    create_layer_of_gates,
)
from orquestra.quantum.utils import compare_unitary


class TestCreateLayerOfGates:
    def test_create_layer_of_gates_not_parametrized(self):
        # Given
        number_of_qubits = 4
        target_circuit = Circuit()
        gate = X
        for i in range(number_of_qubits):
            target_circuit += gate(i)

        # When
        layer_of_x = create_layer_of_gates(number_of_qubits, gate)

        # Then
        assert layer_of_x == target_circuit

    def test_create_layer_of_gates_parametrized(self):
        # Given
        gate = RY
        n_qubits_list = [2, 3, 4, 10]

        for n_qubits in n_qubits_list:
            # Given
            params = np.array([x for x in range(n_qubits)]).reshape(-1, 1)
            target_circuit = Circuit()
            for i in range(n_qubits):
                target_circuit += RY(*params[i])(i)

            u_orquestra = target_circuit.to_unitary()

            # When
            circuit = create_layer_of_gates(n_qubits, gate, params)
            unitary = circuit.to_unitary()

            # Then
            assert compare_unitary(unitary, u_orquestra, tol=1e-10)

    def test_create_layer_of_gates_wrong_num_params(self):
        params = np.ones(3).reshape(-1, 1)
        n_qubits = 2
        with pytest.raises(AssertionError):
            _ = create_layer_of_gates(n_qubits, RY, params)


class TestApplyGateToQubits:
    @pytest.fixture
    def base_circuit(self):
        circuit = Circuit()

        circuit += X(0)
        circuit += X(2)
        circuit += RY(np.pi / 3)(1)

        return circuit

    def test_apply_gate_to_qubits_not_parameterized(self, base_circuit):
        test_gate = Y
        qubit_indices = (1, 5, 4, 3)

        should_circuit = base_circuit
        for qubit in qubit_indices:
            should_circuit += test_gate(qubit)
        should_unitary = should_circuit.to_unitary()

        is_circuit = apply_gate_to_qubits(base_circuit, qubit_indices, test_gate)
        is_unitary = is_circuit.to_unitary()

        assert compare_unitary(should_unitary, is_unitary, tol=1e-10)

    @pytest.mark.parametrize(
        "test_gate,parameters",
        [(RY, np.arange(3).reshape(-1, 1)), (U3, np.arange(3 * 3).reshape(-1, 3))],
    )
    def test_apply_gate_to_qubits_parameterized(
        self, base_circuit, test_gate, parameters
    ):
        qubit_indices = (0, 2, 6)

        should_circuit = base_circuit
        for qubit, parameter in zip(qubit_indices, parameters):
            should_circuit += test_gate(*parameter)(qubit)
        should_unitary = should_circuit.to_unitary()

        is_circuit = apply_gate_to_qubits(
            base_circuit, qubit_indices, test_gate, parameters
        )
        is_unitary = is_circuit.to_unitary()

        assert compare_unitary(should_unitary, is_unitary, tol=1e-10)

    def test_apply_gate_to_qubits_wrong_num_params(self, base_circuit):
        params = np.ones(3).reshape(-1, 1)
        n_qubits = (2, 3)
        with pytest.raises(AssertionError):
            _ = apply_gate_to_qubits(base_circuit, n_qubits, RY, params)

    def test_apply_gate_to_qubits_raises_warning_with_duplicates(self, base_circuit):
        params = np.ones(2).reshape(-1, 1)
        n_qubits = (2, 2, 3)

        with pytest.warns(UserWarning):
            _ = apply_gate_to_qubits(base_circuit, n_qubits, RY, params)



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
import json
from typing import Dict, List, TextIO, Tuple

import numpy as np


class CircuitLayers:
    """A class representing a pattern of circuit layers, consisting of lists,
    each list containing the groups of qubits entangled for each multiqubit
    gate in a particular layer.
    """

    def __init__(self, layers: List[List[Tuple]]):
        """
        Args:
            layers: list of list of tuples, each tuple
            representing a group of qubits that is connected in the layer by a
            multiqubit gate.
        """
        self.layers = layers

    def to_dict(self) -> dict:
        return {"layers": self.layers}

    @classmethod
    def from_dict(cls, data: Dict[str, List[List[Tuple]]]):
        layers = [[tuple(x) for x in layer] for layer in data["layers"]]
        return cls(layers)


def save_circuit_layers(circuit_layers: CircuitLayers, filename: str) -> None:
    """Saves a list of circuit layers to a file.
    Args:
        circuit_layers (circuit.CircuitLayers)
        filename (str): the name of the file
    """

    circuit_layers_dict = circuit_layers.to_dict()
    with open(filename, "w") as f:
        f.write(json.dumps(circuit_layers_dict))


def load_circuit_layers(file: TextIO) -> CircuitLayers:
    """Loads a list of circuit layers from a file.
    Args:
        file (str or file-like object): the name of the file, or a file-like object.

    Returns:
        (circuit.CircuitLayers)
    """

    if isinstance(file, str):
        with open(file, "r") as f:
            data = json.load(f)
    else:
        data = json.load(file)

    return CircuitLayers.from_dict(data)


def save_circuit_ordering(ordering, filename):
    """Saves a circuit ordering (e.g. mapping from spin-orbitals to qubits) to a file.
    Args:
        ordering (list)
        filename (str): the name of the file
    """

    ordering = {"ordering": ordering}
    with open(filename, "w") as f:
        f.write(json.dumps(ordering))


def load_circuit_ordering(file):
    """Loads a circuit ordering (e.g. mapping from spin-orbitals to qubits) to a file.
    Args:
        file (str or file-like object): the name of the file, or a file-like object.

    Returns:
        ordering (list)
    """

    if isinstance(file, str):
        with open(file, "r") as f:
            data = json.load(f)
    else:
        data = json.load(file)

    return data["ordering"]


class CircuitConnectivity(object):
    """A class representing the connectivity of a circuit resulting from qpu
    constraints, consisting of a list of tuples of qubits representing the
    allowed multiqubit gate connections.
    """

    def __init__(self, connections):
        """
        Args:
            connections: list of tuples representing groups of qubits
        """
        self.connectivity = connections

    def to_dict(self):
        return {"connectivity": self.connectivity}

    @classmethod
    def from_dict(cls, data):
        tuples = [tuple(x) for x in data["connectivity"]]
        return cls(tuples)


def save_circuit_connectivity(circuit_connectivity, filename):
    """Saves a circuit connectivity to a file.
    Args:
        circuit_connectivity (orquestra.quantum.circuit.CircuitConnectivity)
        filename (str): the name of the file
    """

    circuit_connectivity = circuit_connectivity.to_dict()
    with open(filename, "w") as f:
        f.write(json.dumps(circuit_connectivity))


def load_circuit_connectivity(file):
    """Loads a circuit connectivity from a file.
    Args:
        file (str or file-like object): the name of the file, or a file-like object.

    Returns:
        (orquestra.quantum.circuit.CircuitConnectivity)
    """

    if isinstance(file, str):
        with open(file, "r") as f:
            data = json.load(f)
    else:
        data = json.load(file)

    return CircuitConnectivity.from_dict(data)


def build_circuit_layers_and_connectivity(
    x_dimension, y_dimension=None, layer_type="nearest-neighbor"
):
    """Function to generate circuit layers for 1-dimensional and 2-dimensional
    arrays of qubits
    Args:
        x_dimension (int): number of qubits per row of the array
        y_dimension (int): number of qubits per column of the array
        layer_type (str): string designating type of layer to be created
    Returns:
        (orquestra.quantum.circuit.CircuitConnectivity,
            orquestra.quantum.circuit.CircuitLayers)
    """
    if layer_type == "sycamore":
        return _build_circuit_layers_and_connectivity_sycamore(x_dimension, y_dimension)
    elif layer_type == "nearest-neighbor":
        return _build_circuit_layers_and_connectivity_nearest_neighbors(x_dimension)
    else:
        ValueError("Layer type {0} is not defined".format(layer_type))


def _build_circuit_layers_and_connectivity_sycamore(x_dimension, y_dimension):
    """Function to generate circuit connectivity and circuit layers
        for 2D quantum processors with sycamore-like connectivity
    Args:
        x_dimension (int): number of qubits per row of the array
        y_dimension (int): number of qubits per column of the array
    Returns:
        (orquestra.quantum.circuit.CircuitConnectivity,
            orquestra.quantum.circuit.CircuitLayers)
    """
    connectivity = []

    # two-colors patterns: patterns
    pattern1 = np.zeros((y_dimension - 1, 2 * x_dimension - 1))
    pattern2 = np.zeros((y_dimension - 1, 2 * x_dimension - 1))

    for m in range(0, y_dimension - 1):
        pattern2[m, :] = m % 2

    py_index = 0
    for m, y_index in enumerate(range(1, y_dimension, 2)):
        row_up = []
        row_down = []
        px_index = 0
        for x_index in range(0, x_dimension):
            node = y_index * x_dimension + x_index
            if x_index == x_dimension - 1:
                row_up.append((node, node - x_dimension))
                pattern1[py_index, px_index] = (x_index + m) % 2
                if y_dimension - 1 > y_index:
                    row_down.append((node, node + x_dimension))
                    pattern1[py_index + 1, px_index] = (x_index + m + 1) % 2
                px_index += 1
            else:
                row_up.extend(
                    [(node, node - x_dimension), (node, node - x_dimension + 1)]
                )
                pattern1[py_index, px_index] = (x_index + m) % 2
                pattern1[py_index, px_index + 1] = (x_index + m) % 2
                if y_dimension - 1 > y_index:
                    row_down.extend(
                        [(node, node + x_dimension), (node, node + x_dimension + 1)]
                    )
                    pattern1[py_index + 1, px_index] = (x_index + m + 1) % 2
                    pattern1[py_index + 1, px_index + 1] = (x_index + m + 1) % 2
                px_index += 2
        py_index += 2

        connectivity.append(row_up)
        connectivity.append(row_down)

    # three-color patterns: masks
    masks = []
    for n in range(0, 2):
        mask1 = pattern1.copy()
        mask2 = pattern2.copy()
        for m in range(mask1.shape[0]):
            mask1[m, (m + n) % 2 : mask1.shape[1] : 2] = -1
            mask2[m, (m + n) % 2 : mask1.shape[1] : 2] = -1
        masks.extend([mask1, mask2])

    # generate layers
    layers = []
    for n, mask in enumerate(masks):
        layer1 = []
        layer2 = []
        for x in range(mask.shape[1]):
            for y in range(mask.shape[0]):
                if mask[y, x] == 0:
                    layer1.append(connectivity[y][x])
                elif mask[y, x] == 1:
                    layer2.append(connectivity[y][x])
        layers.extend([layer1, layer2])

    final_connectivity = []
    for group in connectivity:
        final_connectivity.extend(group)

    return CircuitConnectivity(final_connectivity), CircuitLayers(layers)


def _build_circuit_layers_and_connectivity_nearest_neighbors(n_qubits):
    """Function to generate circuit layers for processors with nearest-neighbor
    connectivity
    Args:
        n_qubits (int): number of qubits in the qubit array
    Returns:
        (orquestra.quantum.circuit.CircuitConnectivity,
            orquestra.quantum.circuit.CircuitLayers)
    """
    even_layer = []
    odd_layer = []
    for index in range(0, n_qubits - 1, 2):
        even_layer.append((index, index + 1))
    for index in range(1, n_qubits - 1, 2):
        odd_layer.append((index, index + 1))
    connectivity = []
    connectivity.extend(even_layer)
    connectivity.extend(odd_layer)
    return CircuitConnectivity(connectivity), CircuitLayers([even_layer, odd_layer])



################################################################################
# © Copyright 2020-2022 Zapata Computing Inc.
################################################################################
"""General-purpose utilities."""
import collections
import json
import os
import sys
from contextlib import contextmanager
from functools import lru_cache
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

import numpy as np
import sympy

from .typing import AnyPath, DumpTarget, LoadSource

RNDSEED = 12345


def convert_dict_to_array(dictionary: dict) -> np.ndarray:
    """Convert a dictionary to a numpy array.

    Args:
        dictionary (dict): the dict containing the data

    Returns:
        array (numpy.array): a numpy array
    """

    array = np.array(dictionary["real"])

    if dictionary.get("imag"):
        array = array + 1j * np.array(dictionary["imag"])

    return array


def convert_array_to_dict(array: np.ndarray) -> dict:
    """Convert a numpy array to a dictionary.

    Args:
        array (numpy.array): a numpy array

    Returns:
        dictionary (dict): the dict containing the data
    """

    dictionary = {}
    if np.iscomplexobj(array):
        dictionary["real"] = array.real.tolist()
        dictionary["imag"] = array.imag.tolist()
    else:
        dictionary["real"] = array.tolist()

    return dictionary


def dec2bin(number: int, length: int) -> List[int]:
    """Converts a decimal number into a binary representation
    of fixed number of bits.

    Args:
        number: (int) the input decimal number
        length: (int) number of bits in the output string

    Returns:
        A list of binary numbers
    """

    if pow(2, length) < number:
        sys.exit(
            "Insufficient number of bits for representing the number {}".format(number)
        )

    bit_str = bin(number)
    bit_str = bit_str[2 : len(bit_str)]  # chop off the first two chars
    bit_string = [int(x) for x in list(bit_str)]
    if len(bit_string) < length:
        len_zeros = length - len(bit_string)
        bit_string = [int(x) for x in list(np.zeros(len_zeros))] + bit_string

    return bit_string


def bin2dec(x: List[int]) -> int:
    """Converts a binary vector to an integer, with the 0-th
    element being the most significant digit.

    Args:
        x: (list) a binary vector

    Returns:
        An integer
    """

    dec = 0
    coeff = 1
    for i in range(len(x)):
        dec = dec + coeff * x[len(x) - 1 - i]
        coeff = coeff * 2
    return dec


# The functions PAULI_X, PAULI_Y, PAULI_Z and IDENTITY below are used for
# generating the generators of the Pauli group, which include Pauli X, Y, Z
# operators as well as identity operator

pauli_x = np.array([[0.0, 1.0], [1.0, 0.0]])
pauli_y = np.array([[0.0, -1.0j], [1.0j, 0.0]])
pauli_z = np.array([[1.0, 0.0], [0.0, -1.0]])
identity = np.array([[1.0, 0.0], [0.0, 1.0]])


def is_identity(u: np.ndarray, tol=1e-15) -> bool:
    """Test if a matrix is identity.

    Args:
        u: np.ndarray
            Matrix to be checked.
        tol: float
            Threshold below which two matrix elements are considered equal.
    """

    dims = np.array(u).shape
    if dims[0] != dims[1]:
        raise Exception("Input matrix is not square.")

    return np.allclose(u, np.eye(u.shape[0]), atol=tol)


def is_unitary(u: np.ndarray, tol=1e-15) -> bool:
    """Test if a matrix is unitary.

    Args:
        u: array
            Matrix to be checked.
        tol: float
            Threshold below which two matrix elements are considered equal.
    """

    dims = np.array(u).shape
    if dims[0] != dims[1]:
        raise Exception("Input matrix is not square.")

    test_matrix = np.dot(np.array(u).T.conj(), u)
    return is_identity(test_matrix, tol)


def compare_unitary(u1: np.ndarray, u2: np.ndarray, tol: float = 1e-15) -> bool:
    """Compares two unitary operators to see if they are equal to within a phase.

    Args:
        u1 (numpy.ndarray): First unitary operator.
        u2 (numpy.ndarray): Second unitary operator.
        tol (float): Threshold below which two matrix elements are considered equal.

    Returns:
        bool: True if the unitaries are equal to within the tolerance, ignoring
            differences in global phase.
    """

    if not is_unitary(u1, tol):
        raise Exception("The first input matrix is not unitary.")
    if not is_unitary(u2, tol):
        raise Exception("The second input matrix is not unitary.")

    test_matrix = np.dot(u1.conj().T, u2)
    phase = test_matrix.item((0, 0)) ** -1
    return is_identity(phase * test_matrix, tol)


def sample_from_probability_distribution(
    probability_distribution: dict, n_samples: int
) -> collections.Counter:
    """
    Samples events from a discrete probability distribution

    Args:
        probabilty_distribution: The discrete probability distribution to be used
        for sampling. This should be a dictionary

        n_samples (int): The number of samples desired

    Returns:
        A dictionary of the outcomes sampled. The key values are the things be sampled
        and values are how many times those things appeared in the sampling
    """
    if isinstance(probability_distribution, dict):
        # Need to do this preprocessing to handle different types of dict keys
        keys_as_array = np.empty(len(probability_distribution), dtype=object)
        keys_as_array[:] = list(probability_distribution.keys())

        result = np.random.choice(
            keys_as_array,
            n_samples,
            replace=True,
            p=list(probability_distribution.values()),
        )
        sampled_dict: collections.Counter = collections.Counter(result)

        return sampled_dict
    else:
        raise RuntimeError(
            "Probability distribution should be a dictionary with key value \
        being the thing being sampled and the value being probability of getting \
        sampled "
        )


def convert_bitstrings_to_tuples(bitstrings: Iterable[str]) -> List[Tuple[int, ...]]:
    """Given the measured bitstrings, convert each bitstring to tuple format

    Args:
        bitstrings (list of strings): the measured bitstrings
    Returns:
        A list of tuples
    """
    measurements = [bitstring_to_tuple(bitstring) for bitstring in bitstrings]
    return measurements


@lru_cache()
def bitstring_to_tuple(bitstring: str) -> Tuple[int, ...]:
    """Given a bitstring, convert it to tuple format

    Args:
        bitstring (string): the measured bitstring
    Returns:
        A tuple of 0s and 1s
    """
    measurement = tuple(int(bit) for bit in bitstring[::-1])
    return measurement


def convert_tuples_to_bitstrings(tuples: List[Tuple[int]]) -> List[str]:
    """Given a set of measurement tuples, convert each to a little endian
    string.

    Args:
        tuples (list of tuples): the measurement tuples
    Returns:
        A list of bitstrings
    """
    bitstrings = [tuple_to_bitstring(tup) for tup in tuples]
    return bitstrings


@lru_cache()
def tuple_to_bitstring(tup: Tuple[int, ...]) -> str:
    """Given a tuple, convert to an equivalent string.

    Args:
        tup (tuple): the measurement tuple
    Returns:
        A string with binary digits
    """
    return "".join(map(str, tup))


class ValueEstimate(float):
    """A class representing a numerical value and its precision corresponding
        to an observable or an objective function

    Args:
        value (np.float): the numerical value or a value that can be converted to float
        precision (np.float): its precision

    Attributes:
        value (np.float): the numerical value
        precision (np.float): its precision
    """

    def __init__(self, value, precision: Optional[float] = None):
        super().__init__()
        self.precision = precision

    def __new__(cls, value, precision=None):
        return super().__new__(cls, value)

    def __eq__(self, other):
        super_eq = super().__eq__(other)
        if super_eq is NotImplemented:
            return super_eq
        return super_eq and self.precision == getattr(other, "precision", None)

    def __ne__(self, other):
        return not self == other

    def __str__(self):
        value_str = super().__str__()
        if self.precision is not None:
            return f"{value_str} ± {self.precision}"
        else:
            return f"{value_str}"

    def to_dict(self):
        """Convert to a dictionary"""

        data = {}
        if type(self).__module__ == np.__name__:
            data["value"] = self.value.item()
        else:
            data["value"] = self

        if type(self.precision).__module__ == np.__name__:
            data["precision"] = self.precision.item()
        else:
            data["precision"] = self.precision

        return data

    @classmethod
    def from_dict(cls, dictionary):
        """Create an ExpectationValues object from a dictionary."""

        value = dictionary["value"]
        if "precision" in dictionary:
            precision = dictionary["precision"]
            return cls(value, precision)
        else:
            return cls(value)


def load_value_estimate(file: LoadSource) -> ValueEstimate:
    """Loads value estimate from a failed.

    Args:
        file (str or file-like object): the name of the file, or a file-like object.

    Returns:
        array (numpy.array): the array
    """

    if isinstance(file, str):
        with open(file, "r") as f:
            data = json.load(f)
    else:
        data = json.load(file)  # type: ignore

    return ValueEstimate.from_dict(data)


def save_value_estimate(value_estimate: ValueEstimate, filename: AnyPath):
    """Saves value estimate to a file.

    Args:
        value_estimate (orquestra.quantum.utils.ValueEstimate): the value estimate
        file (str or file-like object): the name of the file, or a file-like object
    """
    dictionary = value_estimate.to_dict()

    with open(filename, "w") as f:
        f.write(json.dumps(dictionary, indent=2))


def load_list(file: LoadSource) -> List:
    """Load an array from a file.

    Args:
        file (str or file-like object): the name of the file, or a file-like object.

    Returns:
        array (list): the list
    """

    if isinstance(file, str):
        with open(file, "r") as f:
            data = json.load(f)
    else:
        data = json.load(file)  # type: ignore

    return data["list"]


def save_list(array: List, filename: AnyPath):
    """Save expectation values to a file.

    Args:
        array (list): the list to be saved
        file (str or file-like object): the name of the file, or a file-like object
    """
    dictionary: Dict[str, Any] = {}
    dictionary["list"] = array

    with open(filename, "w") as f:
        f.write(json.dumps(dictionary, indent=2))


def save_generic_dict(dictionary: Dict, filename: AnyPath):
    """Save dictionary as json

    Args:
        dictionary (dict): the dict containing the data
    """
    dictionary_stored = {}
    dictionary_stored.update(dictionary)

    with open(filename, "w") as f:
        f.write(json.dumps(dictionary_stored, indent=2))


def create_symbols_map(
    symbols: List[sympy.Symbol], params: np.ndarray
) -> Dict[sympy.Symbol, float]:
    """
    Creates a map to be used for evaluating sympy expressions.

    Args:
        symbols: list of sympy Symbols to be evaluated
        params: numpy array containing numerical value for the symbols
    """
    if len(symbols) != len(params):
        raise (
            ValueError(
                "Length of symbols: {0} doesn't match length of params: {1}".format(
                    len(symbols), len(params)
                )
            )
        )
    return {symbol: param for symbol, param in zip(symbols, params.tolist())}


def save_timing(walltime: float, filename: AnyPath) -> None:
    """
    Saves timing information.

    Args:
        walltime: The execution time.
    """

    with open(filename, "w") as f:
        f.write(json.dumps({"walltime": walltime}))


def save_nmeas_estimate(
    nmeas: float,
    nterms: int,
    filename: AnyPath,
    frame_meas: Optional[np.ndarray] = None,
) -> None:
    """Save an estimate of the number of measurements to a file

    Args:
        nmeas: total number of measurements for epsilon = 1.0
        nterms: number of terms (groups) in the objective function
        frame_meas: A list of the number of measurements per frame for epsilon = 1.0
    """

    data: Dict[str, Any] = {}
    data["K"] = nmeas
    data["nterms"] = nterms
    if frame_meas is not None:
        data["frame_meas"] = convert_array_to_dict(frame_meas)

    with open(filename, "w") as f:
        f.write(json.dumps(data, indent=2))


def load_nmeas_estimate(filename: AnyPath) -> Tuple[float, int, np.ndarray]:
    """Load an estimate of the number of measurements from a file.

    Args:
        filename: the name of the file

    Returns:
        nmeas: number of measurements for epsilon = 1.0
        nterms: number of terms in the hamiltonian
        frame_meas: frame measurements (number of measurements per group)
    """

    with open(filename, "r") as f:
        data = json.load(f)

    frame_meas = convert_dict_to_array(data["frame_meas"])
    K_coeff = data["K"]
    nterms = data["nterms"]

    return K_coeff, nterms, frame_meas


def scale_and_discretize(values: Iterable[float], total: int) -> List[int]:
    """Convert a list of floats to a list of integers such that the total equals
    a given value and the ratios of elements are approximately preserved.

    Args:
        values: The list of floats to be scaled and discretized.
        total: The desired total which the resulting values should sum to.

    Returns:
        A list of integers whose sum is equal to the given total, where the
            ratios of the list elements are approximately equal to the ratios
            of the input list elements.
    """

    # Note: combining the two lines below breaks type checking; see mypy #6040
    value_sum = sum(values)
    scale_factor = total / value_sum

    result = [np.floor(value * scale_factor) for value in values]
    remainders = [
        value * scale_factor - np.floor(value * scale_factor) for value in values
    ]
    indexes_sorted_by_remainder = np.argsort(remainders)[::-1]
    for index in range(int(round(total - sum(result)))):
        result[indexes_sorted_by_remainder[index]] += 1

    result = [int(value) for value in result]

    assert sum(result) == total, "The scaled list does not sum to the desired total."

    return result


def get_ordered_list_of_bitstrings(num_qubits: int) -> List[str]:
    """Create list of binary strings corresponding to 2^num_qubits integers
    and save them in ascending order.

    Args:
        num_qubits: number of binary digits in each bitstring

    Returns:
        The ordered bitstring representations of the integers
    """
    bitstrings = []
    for i in range(2**num_qubits):
        bitstring = "{0:b}".format(i)
        while len(bitstring) < num_qubits:
            bitstring = "0" + bitstring
        bitstrings.append(bitstring)
    return bitstrings


@contextmanager
def ensure_open(path_like: Union[LoadSource, DumpTarget], mode="r", encoding="utf-8"):
    # str | bytes | PathLike | Readable
    if isinstance(path_like, (str, bytes, os.PathLike)):
        with open(path_like, mode, encoding=encoding if "b" not in mode else None) as f:
            yield f
    else:
        # Readable | Writable
        if set(mode).intersection(set("wxa+")) and not path_like.writable():
            raise ValueError(f"File isn't writable, can't ensure mode {mode}")
        yield path_like



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
from typing import Any, Dict, List, Tuple

import rapidjson as json

from ..typing import AnyPath, LoadSource
from ._pauli_operators import PauliRepresentation, PauliSum, PauliTerm


def convert_dict_to_op(dictionary: dict) -> PauliSum:
    """Get a PauliSum from a dictionary.
    Args:
        dictionary: the dictionary representation
    Returns:
        op: the operator
    """
    full_operator = PauliSum()
    for term_dict in dictionary["terms"]:
        operator: List[Tuple[str, int]] = []
        for pauli_op in term_dict["pauli_ops"]:
            operator.append((pauli_op["op"], pauli_op["qubit"]))
        coefficient = term_dict["coefficient"]["real"]
        if term_dict["coefficient"].get("imag"):
            coefficient += 1j * term_dict["coefficient"]["imag"]
        full_operator += PauliTerm.from_iterable(operator, coefficient)

    return full_operator


def convert_op_to_dict(op: PauliRepresentation) -> Dict[str, Any]:
    """Convert a PauliTerm or PauliSum to a dictionary.
    Args:
        op: the operator
    Returns:
        dictionary: the dictionary representation
    """

    dictionary: Dict[str, Any] = {}
    dictionary["terms"] = []
    for term in op.terms:
        term_dict: Dict[str, Any] = {
            "pauli_ops": [{"qubit": op[0], "op": op[1]} for op in term.operations]
        }

        if isinstance(term.coefficient, complex):
            term_dict["coefficient"] = {
                "real": term.coefficient.real,
                "imag": term.coefficient.imag,
            }
        else:
            term_dict["coefficient"] = {"real": term.coefficient.real}

        dictionary["terms"].append(term_dict)

    return dictionary


def save_operator(operator: PauliRepresentation, filename: AnyPath) -> None:
    """Save a qubit operator to file.
    Args:
        operator: the operator to be saved
        filename: the name of the file
    """

    with open(filename, "w") as f:
        f.write(json.dumps(convert_op_to_dict(operator), indent=2))


def load_operator(file: LoadSource) -> PauliSum:
    """Load an operator object from a file.
    Args:
        file: the name of the file, or a file-like object.
    Returns:
        op: the operator.
    """

    if isinstance(file, str):
        with open(file, "r") as f:
            data = json.load(f)
    else:
        data = json.load(file)

    return convert_dict_to_op(data)


def save_operator_set(operator_set: List[PauliSum], filename: AnyPath) -> None:
    """Save a set of qubit operators to a file.

    Args:
        operator_set: a list of QubitOperator to be saved
        file: the name of the file
    """
    dictionary: Dict[str, Any] = {}
    dictionary["operators"] = []
    for operator in operator_set:
        dictionary["operators"].append(convert_op_to_dict(operator))
    with open(filename, "w") as f:
        f.write(json.dumps(dictionary, indent=2))


def load_operator_set(file: LoadSource) -> List[PauliSum]:
    """Load a set of qubit operators from a file.

    Args:
        file: the name of the file, or a file-like object.

    Returns:
        operator_set: a list of QubitOperator objects
    """
    if isinstance(file, str):
        with open(file, "r") as f:
            data = json.load(f)
    else:
        data = json.load(file)

    operator_set = []
    for operator_dict in data["operators"]:
        operator_set.append(convert_dict_to_op(operator_dict))
    return operator_set


def get_pauli_strings(operator: PauliRepresentation) -> List[str]:
    """Convert a operator into a list of Pauli strings.

    Args:
        operator: an operator to be converted

    Returns:
        pauli_strings: list of Pauli strings
    """
    pauli_strings = []
    for term in operator.terms:
        pauli_string = ""
        for pauli in sorted(term.operations):
            pauli_string += pauli[1] + str(pauli[0])
        pauli_strings.append(pauli_string)

    return pauli_strings



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
from abc import abstractmethod
from typing import Iterable, Protocol, Sequence, TypeVar

OperationType = TypeVar("OperationType")


class DecompositionRule(Protocol[OperationType]):
    @abstractmethod
    def production(self, operation: OperationType) -> Iterable[OperationType]:
        """Decompose given operation.

        Implementers should make sure that the result of applying the decomposed
        sequence to any wavefunction is the same as after applying the original
        operation.
        """

    @abstractmethod
    def predicate(self, operation: OperationType) -> bool:
        """Determine if this rule should be applied to given operation."""


def decompose_operation(
    operation: OperationType,
    decomposition_rules: Sequence[DecompositionRule[OperationType]],
):
    if not decomposition_rules:
        return [operation]

    current_rule, *remaining_rules = decomposition_rules

    new_operations_to_decompose = (
        current_rule.production(operation)
        if current_rule.predicate(operation)
        else [operation]
    )
    return [
        decomposed_op
        for op in new_operations_to_decompose
        for decomposed_op in decompose_operation(op, remaining_rules)
    ]


def decompose_operations(
    operations: Iterable[OperationType],
    decomposition_rules: Sequence[DecompositionRule[OperationType]],
):
    return [
        decomposed_op
        for op in operations
        for decomposed_op in decompose_operation(op, decomposition_rules)
    ]



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
import os

import numpy as np

from orquestra.quantum.measurements import (
    ExpectationValues,
    Parities,
    concatenate_expectation_values,
    expectation_values_to_real,
    get_expectation_values_from_parities,
    load_expectation_values,
    save_expectation_values,
)


def remove_file_if_exists(filename):
    try:
        os.remove(filename)
    except OSError:
        pass


def test_expectation_values_io():
    expectation_values = np.array([0.0, 0.0, -1.0])
    correlations = []
    correlations.append(np.array([[1.0, -1.0], [-1.0, 1.0]]))
    correlations.append(np.array([[1.0]]))

    estimator_covariances = []
    estimator_covariances.append(np.array([[0.1, -0.1], [-0.1, 0.1]]))
    estimator_covariances.append(np.array([[0.1]]))

    expectation_values_object = ExpectationValues(
        expectation_values, correlations, estimator_covariances
    )

    save_expectation_values(expectation_values_object, "expectation_values.json")
    expectation_values_object_loaded = load_expectation_values(
        "expectation_values.json"
    )

    assert np.allclose(
        expectation_values_object.values,
        expectation_values_object_loaded.values,
    )
    assert len(expectation_values_object.correlations) == len(
        expectation_values_object_loaded.correlations
    )
    assert len(expectation_values_object.estimator_covariances) == len(
        expectation_values_object_loaded.estimator_covariances
    )
    for i in range(len(expectation_values_object.correlations)):
        assert np.allclose(
            expectation_values_object.correlations[i],
            expectation_values_object_loaded.correlations[i],
        )
    for i in range(len(expectation_values_object.estimator_covariances)):
        assert np.allclose(
            expectation_values_object.estimator_covariances[i],
            expectation_values_object_loaded.estimator_covariances[i],
        )

    remove_file_if_exists("expectation_values.json")


def test_get_expectation_values_from_parities():
    parities = Parities(values=np.array([[18, 50], [120, 113], [75, 26]]))
    expectation_values = get_expectation_values_from_parities(parities)

    assert len(expectation_values.values) == 3
    assert np.isclose(expectation_values.values[0], -0.47058823529411764)
    assert np.isclose(expectation_values.values[1], 0.030042918454935622)
    assert np.isclose(expectation_values.values[2], 0.48514851485148514)

    assert len(expectation_values.estimator_covariances) == 3
    assert np.allclose(
        expectation_values.estimator_covariances[0],
        np.array([[0.014705882352941176]]),
    )
    assert np.allclose(
        expectation_values.estimator_covariances[1], np.array([[0.00428797]])
    )

    assert np.allclose(
        expectation_values.estimator_covariances[2], np.array([[0.0075706]])
    )


def test_expectation_values_to_real():
    # Given
    expectation_values = ExpectationValues(np.array([0.0 + 0.1j, 0.0 + 1e-10j, -1.0]))
    target_expectation_values = ExpectationValues(np.array([0.0, 0.0, -1.0]))

    # When
    real_expectation_values = expectation_values_to_real(expectation_values)

    # Then
    for value in expectation_values.values:
        assert not isinstance(value, complex)
    np.testing.assert_array_equal(
        real_expectation_values.values, target_expectation_values.values
    )


def test_concatenate_expectation_values():
    expectation_values_set = [
        ExpectationValues(np.array([1.0, 2.0])),
        ExpectationValues(np.array([3.0, 4.0])),
    ]

    combined_expectation_values = concatenate_expectation_values(expectation_values_set)
    assert combined_expectation_values.correlations is None
    assert combined_expectation_values.estimator_covariances is None
    assert np.allclose(combined_expectation_values.values, [1.0, 2.0, 3.0, 4.0])


def test_concatenate_expectation_values_with_cov_and_corr():
    expectation_values_set = [
        ExpectationValues(
            np.array([1.0, 2.0]),
            estimator_covariances=[np.array([[0.1, 0.2], [0.3, 0.4]])],
            correlations=[np.array([[-0.1, -0.2], [-0.3, -0.4]])],
        ),
        ExpectationValues(
            np.array([3.0, 4.0]),
            estimator_covariances=[np.array([[0.1]]), np.array([[0.2]])],
            correlations=[np.array([[-0.1]]), np.array([[-0.2]])],
        ),
    ]
    combined_expectation_values = concatenate_expectation_values(expectation_values_set)
    assert len(combined_expectation_values.estimator_covariances) == 3
    assert np.allclose(
        combined_expectation_values.estimator_covariances[0],
        [[0.1, 0.2], [0.3, 0.4]],
    )
    assert np.allclose(combined_expectation_values.estimator_covariances[1], [[0.1]])
    assert np.allclose(combined_expectation_values.estimator_covariances[2], [[0.2]])

    assert len(combined_expectation_values.correlations) == 3
    assert np.allclose(
        combined_expectation_values.correlations[0],
        [[-0.1, -0.2], [-0.3, -0.4]],
    )
    assert np.allclose(combined_expectation_values.correlations[1], [[-0.1]])
    assert np.allclose(combined_expectation_values.correlations[2], [[-0.2]])

    assert np.allclose(combined_expectation_values.values, [1.0, 2.0, 3.0, 4.0])



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
import json
import os
import random

import numpy as np
import pytest
import sympy
from scipy.stats import unitary_group

from orquestra.quantum.utils import (
    RNDSEED,
    ValueEstimate,
    bin2dec,
    compare_unitary,
    convert_array_to_dict,
    convert_dict_to_array,
    convert_tuples_to_bitstrings,
    create_symbols_map,
    dec2bin,
    get_ordered_list_of_bitstrings,
    is_identity,
    is_unitary,
    load_list,
    load_nmeas_estimate,
    load_value_estimate,
    sample_from_probability_distribution,
    save_generic_dict,
    save_list,
    save_nmeas_estimate,
    save_timing,
    save_value_estimate,
    scale_and_discretize,
)


def remove_file_if_exists(filename):
    try:
        os.remove(filename)
    except OSError:
        pass


class TestUtils:
    def test_real_array_conversion(self):
        arr = np.array([[1.0, 2.0], [3.0, 4.0]])
        dictionary = convert_array_to_dict(arr)
        new_arr = convert_dict_to_array(dictionary)
        assert np.allclose(arr, new_arr)

    def test_complex_array_conversion(self):
        arr = np.array([[1.0, 2.0], [3.0, 4.0j]])
        dictionary = convert_array_to_dict(arr)
        new_arr = convert_dict_to_array(dictionary)
        assert np.allclose(arr, new_arr)

    def test_dec_bin_conversion(self):
        integer = random.randint(1, 10**9)
        integer2 = bin2dec(dec2bin(integer, 30))
        assert integer == integer2

    def test_is_identity(self):
        # Given
        matrix1 = np.eye(4)
        random.seed(RNDSEED)
        matrix2 = np.array(
            [[random.uniform(-1, 1) for x in range(0, 4)] for y in range(0, 4)]
        )
        # When/Then
        assert is_identity(matrix1)
        assert not is_identity(matrix2)

    def test_is_unitary(self):
        # Given
        U = unitary_group.rvs(4)
        # When/Then
        assert is_unitary(U)

    def test_compare_unitary(self):
        # Given
        U1 = unitary_group.rvs(4)
        U2 = unitary_group.rvs(4)
        # When/Then
        assert not compare_unitary(U1, U2)

    @pytest.mark.parametrize(
        "distribution, number_of_samples",
        [
            ({"00": 0.5, "11": 0.5}, 1),
            ({"00": 0.5, "11": 0.5}, 10),
            ({"00": 0.5, "11": 0.5}, 17),
            ({"00": 0.5, "11": 0.5}, 177),
            ({"0000": 0.137, "0001": 0.863}, 10),
            ({"0000": 0.137, "0001": 0.863}, 100),
        ],
    )
    def test_sample_from_probability_distribution_gives_correct_number_of_samples(
        self, distribution, number_of_samples
    ):
        counts = sample_from_probability_distribution(distribution, number_of_samples)
        assert sum(counts.values()) == number_of_samples

    def test_convert_bitstrings_to_tuples(self):
        pass

    def test_convert_tuples_to_bitstrings(self):
        pass

    def test_value_estimate_dict_conversion(self):
        pass

    def test_value_estimate_io(self):
        # Given
        value = -1.0
        precision = 0.1
        value_estimate_object = ValueEstimate(value, precision)

        # When
        save_value_estimate(value_estimate_object, "value_estimate.json")
        value_estimate_object_loaded = load_value_estimate("value_estimate.json")

        # Then
        assert value_estimate_object == value_estimate_object_loaded
        assert value_estimate_object.precision == value_estimate_object_loaded.precision

        # Given
        value_estimate_object = ValueEstimate(value)
        # When
        save_value_estimate(value_estimate_object, "value_estimate.json")
        value_estimate_object_loaded = load_value_estimate("value_estimate.json")
        # Then
        assert value_estimate_object == value_estimate_object_loaded
        assert value_estimate_object.precision == value_estimate_object_loaded.precision

        # Given
        value = np.float64(-1.0)
        precision = np.float64(0.1)
        value_estimate_object = ValueEstimate(value, precision)

        # When
        save_value_estimate(value_estimate_object, "value_estimate.json")
        value_estimate_object_loaded = load_value_estimate("value_estimate.json")

        # Then
        assert value_estimate_object == value_estimate_object_loaded
        assert value_estimate_object.precision == value_estimate_object_loaded.precision

        remove_file_if_exists("value_estimate.json")

    def test_value_estimate_to_string(self):
        value = -1.0
        precision = 0.1
        value_estimate = ValueEstimate(value, precision)
        assert str(value_estimate) == f"{value} ± {precision}"

        value_estimate_no_precision = ValueEstimate(value)
        assert str(value_estimate_no_precision) == f"{value}"

    def test_list_io(self):
        # Given
        initial_list = [0.1, 0.3, -0.3]
        # When
        save_list(initial_list, "list.json")
        loaded_list = load_list("list.json")
        # Then
        assert initial_list == loaded_list
        remove_file_if_exists("list.json")

    def test_named_list_io(self):
        # Given
        initial_list = [0.1, 0.3, -0.3]
        # When
        save_list(initial_list, "list.json")
        loaded_list = load_list("list.json")
        # Then
        assert initial_list == loaded_list

    def test_save_generic_dict(self):
        data = {"flavor": "chocolate", "weight": 42}
        save_generic_dict(data, "dict.json")
        with open("dict.json") as f:
            loaded_data = json.load(f)
        for key, value in data.items():
            assert loaded_data[key] == value
        remove_file_if_exists("dict.json")

    def test_create_symbols_map_with_correct_input(self):
        # Given
        symbol_1 = sympy.Symbol("alpha")
        symbol_2 = sympy.Symbol("beta")
        symbols = [symbol_1, symbol_2]
        params = np.array([1, 2])
        target_symbols_map = {symbol_1: 1, symbol_2: 2}

        # When
        symbols_map = create_symbols_map(symbols, params)

        # Then
        assert symbols_map == target_symbols_map

    def test_create_symbols_map_with_incorrect_input(self):
        # Given
        symbol_1 = sympy.Symbol("alpha")
        symbols = [symbol_1]
        params = np.array([1, 2])

        # When/Then
        with pytest.raises(ValueError):
            create_symbols_map(symbols, params)

    def test_save_timing(self):
        walltime = 4.2
        save_timing(walltime, "timing.json")
        with open("timing.json") as f:
            timing = json.load(f)
        assert timing["walltime"] == walltime
        remove_file_if_exists("timing.json")

    def test_save_nmeas_estimate(self):
        K_coeff = 0.5646124437984263
        nterms = 14
        frame_meas = np.array(
            [0.03362557, 0.03362557, 0.03362557, 0.03362557, 0.43011016]
        )
        save_nmeas_estimate(
            nmeas=K_coeff,
            nterms=nterms,
            filename="hamiltonian_analysis.json",
            frame_meas=frame_meas,
        )
        K_coeff_, nterms_, frame_meas_ = load_nmeas_estimate(
            "hamiltonian_analysis.json"
        )
        assert K_coeff == K_coeff_
        assert nterms == nterms_
        assert frame_meas.tolist() == frame_meas_.tolist()
        remove_file_if_exists("hamiltonian_analysis.json")


def test_arithmetic_on_value_estimate_and_float():
    value = 5.1
    estimate = ValueEstimate(value, precision=None)
    other = 3.4

    assert estimate + other == value + other
    assert estimate - other == value - other
    assert estimate * other == value * other
    assert estimate / other == value / other


def test_value_estimate_with_no_precision_is_equivalent_to_its_raw_value():
    value = 6.193
    estimate = ValueEstimate(value)

    # Note that it is not that obvious that this comparison is symmetric, since we
    # override the __eq__ method in ValueEstimate. The same goes about __ne__ method in
    # the next test.
    assert value == estimate
    assert estimate == value


def test_value_estimate_with_specified_precision_is_not_equal_to_its_raw_value():
    value = 6.193
    estimate = ValueEstimate(value, precision=4)

    assert value != estimate
    assert estimate != value


@pytest.mark.parametrize(
    "estimate_1,estimate_2,expected_result",
    [
        (ValueEstimate(14.1), ValueEstimate(14.1), True),
        (ValueEstimate(12.3, 3), ValueEstimate(12.3, 3), True),
        (ValueEstimate(14.1, 5), ValueEstimate(14.1, 4), False),
        (ValueEstimate(2.5, 3), ValueEstimate(2.5), False),
        (ValueEstimate(0.15, 3), ValueEstimate(1.1, 3), False),
    ],
)
def test_two_value_estimates_are_equal_iff_their_values_and_precisions_are_equal(
    estimate_1, estimate_2, expected_result
):
    assert (estimate_1 == estimate_2) == expected_result


@pytest.mark.parametrize(
    "estimate", [ValueEstimate(2.0), ValueEstimate(5.0, precision=1e-5)]
)
@pytest.mark.parametrize("other_obj", ["test-string", {"foo": 5, "bar": 10}, [1, 2, 3]])
def test_value_estimate_is_not_equivalent_to_an_object_of_non_numeric_type(
    estimate, other_obj
):
    assert estimate != other_obj


@pytest.mark.parametrize(
    "values,total,expected_result",
    [
        ([0.5, 0.3, 0.2], 9, [4, 3, 2]),
        ([0.5, 0.3, 0.2], 10, [5, 3, 2]),
        ([0.5, 0.3, 0.2], 11, [6, 3, 2]),
        ([0.5, 0.3, 0.2], 2, [1, 1, 0]),
    ],
)
def test_scale_and_discretize(values, total, expected_result):
    assert scale_and_discretize(values, total) == expected_result


@pytest.mark.parametrize("num_qubits", [2, 3, 5, 10])
def test_ordered_bitstring(num_qubits):
    bitstrings = get_ordered_list_of_bitstrings(num_qubits)
    expected_bitstrings = convert_tuples_to_bitstrings(
        [tuple(dec2bin(integer, num_qubits)) for integer in range(2**num_qubits)]
    )
    assert np.all(expected_bitstrings == bitstrings)
    assert np.all([len(bitstring) == num_qubits for bitstring in bitstrings])
